"""
Methods

assertEqual(a, b)
assertNotEqual(a, b)
assertTrue(x)
assertFalse(x)
assertIs(a, b)
assertIsNot(a, b)
assertIsNone(x)
assertIsNotNone(x)
assertIn(a, b)
assertNotIn(a, b)
assertIsInstance(a, b)
assertNotIsInstance(a, b)
"""
import unittest
import pandas as pd
import numpy as np
import trikit



# Ensure sample datasets have been properly loaded ----------------------------

class DatasetsTestCase(unittest.TestCase):

    def setUp(self):
        self.dactual = {
            "raa"         :160987,
            "ta83"        :34358090,
            "autoliab"    :2197134,
            "glre"        :56531053,
            "singinjury"  :11026482.299999999,
            "singproperty":25101206.26,
            "lrdb"        :10178930,
            "amw09"       :92741342.0,
            }

    def test_raa(self):
        self.assertEqual(
            trikit.load("raa").value.sum(), self.dactual["raa"],
            "Issue detected with raa sample dataset."
            )

    def test_ta83(self):
        self.assertEqual(
            trikit.load("ta83").value.sum(), self.dactual["ta83"],
            "Issue detected with ta83 sample dataset."
            )

    def test_autoliab(self):
        self.assertEqual(
            trikit.load("autoliab").value.sum(), self.dactual["autoliab"],
            "Issue detected with autoliab sample dataset."
            )

    def test_glre(self):
        self.assertEqual(
            trikit.load("glre").value.sum(), self.dactual["glre"],
            "Issue detected with glre sample dataset."
            )

    def test_singinjury(self):
        self.assertEqual(
            trikit.load("singinjury").value.sum(), self.dactual["singinjury"],
            "Issue detected with singinjury sample dataset."
            )

    def test_singproperty(self):
        self.assertEqual(
            trikit.load("singproperty").value.sum(), self.dactual["singproperty"],
            "Issue detected with singproperty sample dataset."
            )

    def test_raa(self):
        self.assertEqual(
            trikit.load("raa").value.sum(), self.dactual["raa"],
            "Issue detected with raa sample dataset."
            )

    def test_amw09(self):
        self.assertEqual(
            trikit.load("amw09").value.sum(), self.dactual["amw09"],
            "Issue detected with amw09 sample dataset."
            )

    def test_lrdb(self):
        self.assertEqual(
            trikit.load_lrdb().value.sum(), self.dactual["lrdb"],
            "Issue detected with lrdb sample dataset."
            )


if __name__ == "__main__":

    unittest.main()


"""
This module contains the class definition of ``BaseChainLadder``.
"""
from collections.abc import Sequence
import functools
import warnings
import pandas as pd
import numpy as np



class BaseChainLadder:
    """
    From the Casualty Actuarial Society's *Estimating Unpaid Claims Using
    Basic Techniques* Version 3 (Friedland, Jacqueline - 2010), the
    development method ('Chain Ladder') consists of seven basic steps:

    1. Compile claims data in a development triangle.
    2. Calculate age-to-age factors.
    3. Calculate averages of the age-to-age factors.
    4. Select claim development factors.
    5. Select tail factor.
    6. Calculate cumulative claims.
    7. Project ultimate claims.

    The BaseChainLadder class encapsulates logic to perform steps 1-7.

    Parameters
    ----------
    cumtri: trikit.triangle.CumTriangle
        A cumulative triangle instance.

    References
    ----------
    1. Friedland, J., *Estimating Unpaid Claims Using Basic Techniques*,
       Casualty Actuarial Society, 2010.
    """
    def __init__(self, cumtri):
        """
        Generate point estimates for outstanding claim liabilities at
        ultimate for each origin year and in aggregate. The
        BaseChainLadder class exposes no functionality to estimate
        variability around the point estimates at ultimate.

        Parameters
        ----------
        cumtri: triangle._CumTriangle
            A cumulative.CumTriangle instance.
        """
        self.tri = cumtri


    def __call__(self, sel="all-weighted", tail=1.0):
        """
        Compile a summary of ultimate and reserve estimates resulting from
        the application of the development technique. Returned object is an
        instance of ``BaseChainLadderResult``, which exposes a ``summary``
        attribute, a DataFrame with the following fields:

            * index: Origin period.

            * maturity: The age of the associated origin period in terms of
              development period duration.

            * cldf: Cumulative loss development factors.

            * emergence: 1 / cldf.

            * latest: The latest diagonal from the cumulative triangle instance.

            * ultimate: Projected ultimates. Computed as latest * cldf.

            * reserve: Chain ladder reserve estimates. Computed as
              ultimate - latest.

        Parameters
        ----------
       sel: str, pd.Series or array_like
            If ``sel`` is a string, the specified loss development patterns will be
            the associated entry from ``self.tri.a2a_avgs``.
            If ``sel`` is array_like, values will be used in place of loss development
            factors computed from the traingle directly. For a triangle with n development
            periods, ``sel`` should be array_like with length n - 1.
            Defaults to "all-weighted".

        tail: float
            Tail factor. Defaults to 1.0.

        Returns
        -------
        BaseChainLadderResult
        """
        if isinstance(sel, str):
            ldfs = self._ldfs(sel=sel, tail=tail)

        else:
            if isinstance(sel, pd.Series):
                # Check whether sel has the appropriate length.
                if sel.index.size != (self.tri.devp.size - 1):
                    raise ValueError(
                        "sel has {} values, LDF overrides require {}.".format(
                            sel.size, self.tri.devp.size - 1
                            )
                        )

                # Append tail factor to sel.
                increment = np.unique(sel.index[1:] - sel.index[:-1])[0]
                sel.loc[sel.index.max() + increment] = tail

            elif isinstance(sel, (Sequence, np.ndarray)):
                sel = np.asarray(sel, dtype=float)
                if len(sel) != len(self.tri.devp) - 1:
                    if sel.size == (self.tri.devp.size - 1):
                        raise ValueError(
                            "sel has {} values, LDF overrides require at least {}.".format(
                                sel.size, self.tri.devp.size - 1
                                )
                            )
                # Append sel with tail.
                sel = np.append(sel, tail)

            # Coerce sel to pd.Series.
            ldfs = pd.Series(sel, index=self.tri.devp, dtype=float)

        cldfs = self._cldfs(ldfs=ldfs)
        ultimates = self._ultimates(cldfs=cldfs)
        reserves = self._reserves(ultimates=ultimates)
        maturity = self.tri.maturity.astype(str)
        latest = self.tri.latest_by_origin
        trisqrd = self._trisqrd(ldfs=ldfs)

        # Compile chain ladder point estimate summary.
        dfmatur = maturity.to_frame().reset_index(drop=False).rename({"index": "origin"}, axis=1)
        dfcldfs = cldfs.to_frame().reset_index(drop=False).rename({"index": "maturity"}, axis=1)
        dfcldfs["maturity"] = dfcldfs["maturity"].astype(str)
        dfcldfs["emergence"] = 1 / dfcldfs["cldf"]
        dfsumm = dfmatur.merge(dfcldfs, on=["maturity"], how="left").set_index("origin")
        dfsumm.index.name = None
        dflatest = latest.to_frame().rename({"latest_by_origin": "latest"}, axis=1)
        dfsumm = functools.reduce(
            lambda df1, df2: df1.join(df2),
            (dflatest, ultimates.to_frame(), reserves.to_frame()), dfsumm
            )

        dfsumm.loc["total"] = dfsumm.sum()
        dfsumm.loc["total", "maturity"] = ""
        dfsumm.loc["total", ["cldf", "emergence"]] = np.NaN

        cl_result = BaseChainLadderResult(
            summary=dfsumm, tri=self.tri, sel=sel, ldfs=ldfs, tail=tail, trisqrd=trisqrd
            )

        return(cl_result)


    def _ldfs(self, sel="all-weighted", tail=1.0):
        """
        Lookup loss development factors corresponding to ``sel``.

        Parameters
        ----------
        sel: str
            The ldf average to select from ``triangle._CumTriangle.a2a_avgs``.
            Defaults to "all-weighted".

        tail: float
            Tail factor. Defaults to 1.0.

        Returns
        -------
        pd.Series
        """
        # Determine index for tail factor.
        ldfs = self.tri.a2a_avgs().loc[sel]
        increment = np.unique(ldfs.index[1:] - ldfs.index[:-1])[0]
        ldfs.loc[ldfs.index.max() + increment] = tail
        return(pd.Series(ldfs, name="ldf").sort_index())


    def _cldfs(self, ldfs):
        """
        Calculate cumulative loss development factors by successive
        multiplication beginning with the tail factor and the oldest
        age-to-age factor. The cumulative claim development factor projects
        the total growth over the remaining valuations. Cumulative claim
        development factors are also known as "Age-to-Ultimate Factors"
        or "Claim Development Factors to Ultimate".

        Parameters
        ----------
        ldfs: pd.Series
            Selected ldfs, typically the output of calling ``self._ldfs``.

        Returns
        -------
        pd.Series
        """
        cldfs = np.cumprod(ldfs.values[::-1])[::-1]
        cldfs = pd.Series(data=cldfs, index=ldfs.index.values, name="cldf")
        return(cldfs.astype(float).sort_index())


    def _ultimates(self, cldfs):
        """
        Ultimate claims are equal to the product of the latest valuation of
        losses (the amount along latest diagonal of any ``_CumTriangle``
        instance) and the appropriate cldf/age-to-ultimate factor. We
        determine the appropriate age-to-ultimate factor based on the age
        of each origin year relative to the evaluation date.

        Parameters
        ----------
        cldfs: pd.Series
            Cumulative loss development factors, conventionally obtained
            via BaseChainLadder's ``_cldfs`` method.

        Returns
        -------
        pd.Series
        """
        ultimates = pd.Series(
            data=self.tri.latest_by_origin.values * cldfs.values[::-1],
            index=self.tri.index, name="ultimate"
            )
        return(ultimates.astype(float).sort_index())


    def _reserves(self, ultimates):
        """
        Return IBNR/reserve estimates by origin and in aggregate. Represents
        the difference between ultimate projections for each origin period
        and the latest cumulative value.
        Since outstanding claim liabilities can be referred to differently
        based on the type of losses represented in the triangle ("ibnr" if
        reported/incurred, "unpaid" if paid losses), we use the general term
        "reserve" to represent the difference between ultimate projections
        and latest cumulative value by origin and in total.

        Parameters
        ----------
        ultimates: pd.Series
            Estimated ultimate losses, conventionally obtained from
            BaseChainLadder's ``_ultimates`` method.

        Returns
        -------
        pd.Series
        """
        reserves = pd.Series(
            data=ultimates - self.tri.latest_by_origin,
            index=self.tri.index, name='reserve')
        return(reserves.astype(float).sort_index())


    def _trisqrd(self, ldfs):
        """
        Project claims growth for each future development period. Returns a
        DataFrame of loss projections for each subsequent development period
        for each origin period. Populates the triangle's lower-right or
        southeast portion (i.e., the result of "squaring the triangle").

        Parameters
        ----------
        ldfs: pd.Series
            Selected ldfs, typically the output of calling ``self._ldfs``.

        Returns
        -------
        pd.DataFrame
        """
        trisqrd = self.tri.copy(deep=True)
        rposf = self.tri.index.size
        clvi = self.tri.clvi["row_offset"]
        for i in enumerate(trisqrd.columns[1:], start=1):
            ii, devp = i[0], i[1]
            ildf, rposi = ldfs.values[ii - 1], clvi[devp] + 1
            trisqrd.iloc[rposi:rposf, ii] = trisqrd.iloc[rposi:rposf, ii - 1] * ildf
        # Multiply right-most column by tail factor.
        max_devp = trisqrd.columns[-1]
        trisqrd["ultimate"] = trisqrd.loc[:, max_devp].values * ldfs.values[-1]
        return(trisqrd.astype(float).sort_index())



class BaseChainLadderResult:
    """
    Container object for BaseChainLadder output.

    Parameters
    ----------
    summary: pd.DataFrame
        Chain Ladder summary compilation.

    tri: trikit.triangle._CumTriangle
        A cumulative triangle instance.

    sel: str or array_like
        Reference to loss development selection. If ldf overrides are
        utilized, ``sel`` will be identical to ``ldfs``.

    ldfs: pd.Series
        Loss development factors.

    tail: float
        Tail factor. Defaults to 1.0.

    trisqrd: pd.DataFrame
        Projected claims growth for each future development period.
   """
    def __init__(self, summary, tri, sel, ldfs, tail, trisqrd):

        self.emergence = summary["emergence"]
        self.ultimate = summary["ultimate"]
        self.maturity = summary["maturity"]
        self.reserve = summary["reserve"]
        self.latest = summary["latest"]
        self.cldfs = summary["cldf"]
        self.summary = summary
        self.trisqrd = trisqrd
        self.ldfs = ldfs
        self.tail = tail
        self.sel = sel
        self.tri = tri

        self._markers = ["o", "v", "^", "s", "8", "p", "D", "d", "h"]

        self._summspecs = {
            "ultimate": "{:,.0f}".format, "reserve": "{:,.0f}".format,
            "latest": "{:,.0f}".format, "cldf": "{:.5f}".format,
            "emergence": "{:.5f}".format,
            }


    @staticmethod
    def _get_yticks(x):
        """
        Determine y axis tick labels for a given maximum loss amount x.
        Return tuple of tick values and ticklabels.

        Parameters
        ----------
        x: float
            Maximum value for a given origin period.

        Returns
        -------
        tuple of ndarrays
        """
        ref_divs = np.power(10, np.arange(10))
        div_index = np.where((x / ref_divs) > 1)[0].max()
        x_div = ref_divs[div_index]

        # Find upper limit for y-axis given origin_max_val.
        yuls_seq = x_div * np.arange(1, 11)
        x_yuls = yuls_seq - x
        yul = yuls_seq[np.where(x_yuls > 0)[0].min()]
        y_ticks = np.linspace(0, yul, num=5)
        y_ticklabels = np.asarray(["{:,.0f}".format(ii) for ii in y_ticks])
        return(y_ticks, y_ticklabels)


    def _data_transform(self):
        """
        Transform dataset for use in FacetGrid plot by origin exhibting chain
        ladder reserve estimates.

        Returns
        -------
        pd.DataFrame
        """
        trisqrd = self.trisqrd.reset_index(drop=False).rename({"index": "origin"}, axis=1)
        df0 = pd.melt(trisqrd, id_vars=["origin"], var_name="dev", value_name="value")

        # Combine df0 with latest cumulative loss by origin period.
        df0 = df0.merge(
            self.latest.reset_index(drop=False).rename({"index": "origin"}, axis=1),
            on="origin", how="left"
            )

        dfult = df0[df0["dev"] == "ultimate"].copy()
        dev_increment = np.unique(self.ldfs.index[1:] - self.ldfs.index[:-1])[0]
        dfult["dev"] = self.ldfs.index.max() + dev_increment
        dfult["rectype"] = "forecast"
        df0 = df0[df0["dev"] != "ultimate"].reset_index(drop=True)

        # Create tabular dataset based on tri.triind. Rows wiith 0s represent
        # actuals, rows with 1 represent forecasts.
        df1 = self.tri.triind.reset_index(drop=False).rename({"index": "origin"}, axis=1)
        df1 = pd.melt(df1, id_vars=["origin"], var_name="dev", value_name="value")
        df1["value"] = df1["value"].map(lambda v: 1 if v == 0 else 0)
        df1 = df1[~np.isnan(df1["value"])].rename(
            {"value": "actual_ind"}, axis=1).reset_index(drop=True)

        # Combine df0 and df1 into a single DataFrame, then perform cleanup
        # actions for cases in which df0 has more records than df1.
        df = pd.merge(df0, df1, on=["origin", "dev"], how="left", sort=False)

        # Bind reference to maximum dev period for each origin.
        dfma = df[df["actual_ind"] == 1].groupby(
            ["origin"])["dev"].max().to_frame().reset_index(drop=False).rename(
            {"index": "origin", "dev": "max_actual"}, axis=1
            )
        df = pd.merge(df, dfma, on="origin", how="left", sort=False)
        df["incl_actual"] = df["actual_ind"].map(lambda v: 1 if v == 1 else 0)
        df["incl_pred"] = df.apply(
            lambda rec: 1 if (rec.actual_ind == 0 or rec.dev == rec.max_actual) else 0,
            axis=1
            )

        # Split data into actual and pred cohorts, then recombine. Note that
        # the latest cumulative loss by origin intentionally appears in both
        # datasets.
        dfact = df[df["incl_actual"] == 1][["origin", "dev", "value", "latest"]]
        dfact["rectype"] = "actual"
        dfpred = df[df["incl_pred"] == 1][["origin", "dev", "value", "latest"]]
        dfpred["rectype"] = "forecast"

        # Create total DataFrame, representing losses across all origin periods
        # by development period and at ultimate.
        dftotal = pd.concat([
            dfpred.groupby(["dev", "rectype"], as_index=False)[["value", "latest"]].sum(),
            dfult.groupby(["dev", "rectype"], as_index=False)[["value", "latest"]].sum()
            ])

        # Combine dfact, dfpred, dfult and dftotal.
        dftotal["origin"] = "total"
        dfall = pd.concat([dfact, dfpred, dfult, dftotal]).reset_index(drop=True).rename(
            {"value": "loss"}, axis=1
            )


        # Add origin index column sort origin columns, which is of type object
        # # since adding "total".
        dfall["dev"] = dfall["dev"].astype(int)
        origin_vals = sorted([int(ii) for ii in dfall["origin"].unique() if ii != "total"])
        dindex = {jj: ii for ii, jj in enumerate(origin_vals)}
        dindex.update({"total": max(dindex.values()) + 1})
        dfall["origin_index"] = dfall["origin"].map(dindex)

        # Add reserve column, defined as value - latest when rectype=="forecast",
        # otherwise 0.
        dfall["reserve"] = dfall.apply(
            lambda rec: rec.loss - rec.latest if rec.rectype == "forecast" else 0,
            axis=1
            )
        column_order = ["origin_index", "origin", "dev", "loss", "reserve", "rectype"]
        return(dfall[column_order].reset_index(drop=True))


    def plot(self, actuals_color="#334488", forecasts_color="#FFFFFF", axes_style="darkgrid",
             context="notebook", col_wrap=4, hue_kws=None, exhibit_path=None, **kwargs):
        """
        Visualize actual losses along with projected chain ladder development.

        Parameters
        ----------
        actuals_color: str
            A color name or hexidecimal code used to represent actual
            observations. Defaults to "#00264C".

        forecasts_color: str
            A color name or hexidecimal code used to represent forecast
            observations. Defaults to "#FFFFFF".

        axes_style: str
            Aesthetic style of plots. Defaults to "darkgrid". Other options
            include: {whitegrid, dark, white, ticks}.

        context: str
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Defaults to "notebook". Additional options include
            {"paper", "talk", "poster"}.

        col_wrap: int
            The maximum number of origin period axes to have on a single row
            of the resulting FacetGrid. Defaults to 5.

        hue_kws: dictionary of param:list of values mapping
            Other keyword arguments to insert into the plotting call to let
            other plot attributes vary across levels of the hue variable
            (e.g. the markers in a scatterplot). Each list of values should
            have length 2, with each index representing aesthetic
            overrides for forecasts and actuals respectively. Defaults to
            ``None``.

        exhibit_path: str
            Path to which exhibit should be written. If None, exhibit will be
            rendered via ``plt.show()``.

        kwargs: dict
            Additional styling options for scatter points. This can override
            default values for ``plt.plot`` objects. For a demonstration,
            See the Examples section.


        Examples
        --------
        Demonstration of passing a dictionary of plot properties in order
        to update the scatter size and marker::

            In [1]: import trikit
            In [2]: tri = trikit.load(dataset="raa", tri_type="cum")
            In [3]: cl = tri.base_cl(sel="all-weighted", tail=1.005)
            In [4]: kwds = dict(marker="s", markersize=6)
            In [5]: cl.plot(**kwds)

        """
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set_context(context)

        data = self._data_transform()

        with sns.axes_style(axes_style):

            huekwargs = dict(
                marker=["o", "o"], markersize=[6, 6],
                color=["#000000", "#000000"], fillstyle=["full", "full"],
                markerfacecolor=[forecasts_color, actuals_color],
                markeredgecolor=["#000000", "#000000"],
                markeredgewidth=[.50, .50], linestyle=["-", "-"],
                linewidth=[.475, .475],
                )

            if hue_kws is not None:
                # Determine whether the length of each element of hue_kws is 4.
                if all(len(hue_kws[i]) == 4 for i in hue_kws):
                    huekwargs.update(hue_kws)
                else:
                    warnings.warn("hue_kws overrides not correct length - Ignoring.")

            grid = sns.FacetGrid(
                data, col="origin", hue="rectype", hue_kws=huekwargs,
                col_wrap=col_wrap, margin_titles=False, despine=True, sharex=False,
                sharey=False, hue_order=["forecast", "actual"]
                )

            devp_xticks = np.sort(data.dev.unique())
            devp_xticks_str = [
                str(ii) if ii != devp_xticks.max() else "ult" for ii in devp_xticks
                ]
            grid.set(xticks=devp_xticks)
            grid.set_xticklabels(devp_xticks_str, size=7)
            origin_order = data[["origin_index", "origin"]].drop_duplicates().sort_values(
                "origin_index").origin.values

            with warnings.catch_warnings():

                warnings.simplefilter("ignore")

                for origin, ax_ii in zip(origin_order, grid.axes):

                    legend = ax_ii.legend(
                        loc="lower right", fontsize="x-small", frameon=True,
                        fancybox=True, shadow=False, edgecolor="#909090",
                        framealpha=1, markerfirst=True,
                        )
                    legend.get_frame().set_facecolor("#FFFFFF")

                    # For given origin, determine optimal 5-point tick labels.
                    origin_max_val = data[data.origin == origin].loss.max()
                    y_ticks, y_ticklabels = self._get_yticks(origin_max_val)
                    ax_ii.set_yticks(y_ticks)
                    ax_ii.set_yticklabels(y_ticklabels, size=7)
                    ax_ii.annotate(
                        origin, xy=(.075, .90), xytext=(.075, .90), xycoords='axes fraction',
                        textcoords='axes fraction', fontsize=9, rotation=0, color="#000000",
                        )
                    ax_ii.set_title("")
                    ax_ii.set_xlabel("")
                    ax_ii.set_ylabel("")

                    # Draw border around each facet.
                    for _, spine in ax_ii.spines.items():
                        spine.set(visible=True, color="#000000", linewidth=.50)

            if exhibit_path is not None:
                plt.savefig(exhibit_path)
            else:
                plt.show()


    def __str__(self):
        return(self.summary.to_string(formatters=self._summspecs))


    def __repr__(self):
        return(self.summary.to_string(formatters=self._summspecs))





class BaseRangeEstimator(BaseChainLadder):

    @staticmethod
    def _qtls_formatter(q, two_sided=False):
        """
        Return array_like of formatted quantiles.

        Parameters
        ----------
        q: array_like of float or float
            Quantile or sequence of quantiles to compute, which must be
            between 0 and 1 inclusive.

        two_sided: bool
            Whether the two_sided interval should be included in summary
            output. For example, if ``two_sided==True`` and ``q=.95``, then
            the 2.5th and 97.5th quantiles of the estimated reserve
            distribution will be returned [(1 - .95) / 2, (1 + .95) / 2]. When
            False, only the specified quantile(s) will be computed. Defaults
            to False.

        Returns
        -------
        tuple of list
        """
        qtls = np.asarray([q] if isinstance(q, (float, int)) else q)
        if np.all(np.logical_and(qtls <= 1, qtls >= 0)):
            if two_sided:
                qtls = np.sort(np.unique(np.append((1 - qtls) / 2., (1 + qtls) / 2.)))
            else:
                qtls = np.sort(np.unique(qtls))
        else:
            raise ValueError("Values for quantiles must fall between [0, 1].")
        qtlhdrs = [
            "{:.5f}".format(ii).rstrip("0").rstrip(".") + "%" for ii in 100 * qtls
        ]
        return(qtls, qtlhdrs)



class BaseRangeEstimatorResult(BaseChainLadderResult):

    def __init__(self, summary, tri, ldfs, tail, trisqrd, process_error, parameter_error):
        """
        Container class for reserve estimators which quantify reserve variability.
        """
        super().__init__(summary=summary, tri=tri, ldfs=ldfs, tail=tail, trisqrd=trisqrd, sel=None)

        self.parameter_error = parameter_error
        self.process_error = process_error
        self.std_error = summary["std_error"]
        self.mse = summary["std_error"]**2
        self.cv = summary["cv"]

        # Quantile suffix for plot method annotations.
        self.dsuffix = {
            "0": "th", "1": "st", "2": "nd", "3": "rd", "4": "th", "5": "th", "6": "th",
            "7": "th", "8": "th", "9": "th",
            }

        self._summspecs.update({"std_error": "{:,.0f}".format, "cv": "{:.3f}".format})



    def _qtls_formatter(self, q):
        """
        Return array_like of actual and formatted quantiles.

        Parameters
        ----------
        q: array_like of float or float
            Quantile or sequence of quantiles to compute, which must be
            between 0 and 1 inclusive.

        Returns
        -------
        tuple of list
        """
        qtls = np.asarray([q] if isinstance(q, (float, int)) else q)
        if np.all(np.logical_and(qtls <= 1, qtls >= 0)):
            qtls = np.sort(np.unique(qtls))
        else:
            raise ValueError("Values for quantiles must fall between [0, 1].")

        qtlhdrs = [
            "{:.5f}".format(ii).rstrip("0").rstrip(".") for ii in 100 * qtls
            ]
        qtlhdrs = [
            ii + "th" if "." in ii else ii + self.dsuffix[ii[-1]] for ii in qtlhdrs
            ]
        return(qtls.tolist(), qtlhdrs)


    def get_quantiles(self):
        """
        Estimator specific routine to produce quantiles of estimated reserve distribution.
        """
        pass



"""
This module contains the definitions of both the ``IncrTriangle`` and
``CumTriangle`` classes. Users should avoid instantiating ``IncrTriangle``
or ``CumTriangle`` instances directly; rather the dataset and triangle
arguments should be passed to ``totri``, which will return either an
instance of ``CumTriangle`` or ``IncrTriangle``, depending on the argument
specified for ``type_``.
"""
import itertools
import numpy as np
import pandas as pd
from scipy import stats
from .estimators.base import BaseChainLadder
from .estimators.bootstrap import BootstrapChainLadder
from .estimators.mack import MackChainLadder
from .estimators.glm import GLMEstimator


class _BaseTriangle(pd.DataFrame):
    """
    Transforms ``data`` into a triangle instance.

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be transformed into a ``_BaseTriangle`` instance.
        ``data`` must be tabular loss data with at minimum columns
        representing the origin/acident year, the development
        period and the actual loss amount, given by ``origin``, ``dev``
        and ``value`` arguments.

    origin: str
        The fieldname in ``data`` representing origin period.

    dev: str
        The fieldname in ``data`` representing development period.

    value: str
        The fieldname in ``data`` representing loss amounts.
    """
    def __init__(self, data, origin=None, dev=None, value=None):

        self._validate(data, origin=origin, dev=dev, value=value)
        origin_ = "origin" if origin is None else origin
        dev_ = "dev" if dev is None else dev
        value_ = "value" if value is None else value

        # data2 = data.copy(deep=True)
        data2 = self._neg_handler(data, dev_, value_)
        data2 = data2[[origin_, dev_, value_]]
        data2 = data2.groupby([origin_, dev_], as_index=False).sum()
        data2 = data2.sort_values(by=[origin_, dev_])
        tri = data2.pivot(index=origin_, columns=dev_).rename_axis(None)
        tri.columns = tri.columns.droplevel(0)

        # Force all triangle cells to be of type np.floa
        tri = tri.astype({kk: float for kk in tri.columns})
        tri.columns.name = None

        super().__init__(tri)

        self.origin = origin_
        self.value = value_
        self.dev = dev_

        # Properties.
        self._latest_by_origin = None
        self._latest_by_devp = None
        self._nbr_cells = None
        self._maturity = None
        self._triind = None
        self._devp = None
        self._latest = None
        self._origins = None
        self._rlvi = None
        self._clvi = None
        self._dof = None


    @staticmethod
    def _validate(data, origin=None, dev=None, value=None):
        """
        Ensure data has requisite columns.

        Parameters
        ----------
        data: pd.DataFrame
            Initial dataset to be coerced to triangle.

        origin: str
            The fieldname in ``data`` representing origin period.

        dev: str
            The fieldname in ``data`` representing development period.

        value: str
            The fieldname in ``data`` representing loss amounts.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError("`data` must be an instance of pd.DataFrame.")

        origin_ = "origin" if origin is None else origin
        if origin_ not in data.columns:
            raise AttributeError("`{}` not present in data.".format(origin_))

        dev_ = "dev" if dev is None else dev
        if dev_ not in data.columns:
            raise AttributeError("`{}` not present in data.".format(dev_))

        value_ = "value" if value is None else value
        if value_ not in data.columns:
            raise AttributeError("`{}` not present in data.".format(value_))


    @staticmethod
    def _neg_handler(data, dev, value):
        """
        Convert any first development period negative values to 1.0.
        """
        min_devp = data[dev].min()
        data[value] = data.apply(lambda rec: 1 if
            rec[dev] == min_devp and rec[value] <= 0 else rec[value], axis=1)
        return(data)


    @property
    def nbr_cells(self):
        """
        Return the number of non-NaN cells.

        Returns
        -------
        int
        """
        if self._nbr_cells is None:
            self._nbr_cells = self.count().sum()
        return(self._nbr_cells)


    @property
    def triind(self):
        """
        Table indicating forecast cells with 1, actual data with 0.

        Returns
        -------
        pd.DataFrame
        """
        if self._triind is None:
            self._triind = pd.DataFrame(columns=self.columns, index=self.index)
            self._triind.iloc[:, :] = 0
            for devp in self.clvi.index:
                last_actual_origin = self.clvi[self.clvi.index == devp].origin.values[0]
                last_actual_offset = self.clvi[self.clvi.origin == last_actual_origin].row_offset.values[0]
                self._triind.iloc[(last_actual_offset + 1):, self.columns.get_loc(devp)] = 1
        return(self._triind)


    @property
    def rlvi(self):
        """
        Determine the last valid index by origin.

        Returns
        -------
        pd.DataFrame
        """
        if self._rlvi is None:
            self._rlvi = pd.DataFrame({
                "dev": self.apply(
                    lambda x: x.last_valid_index(), axis=1).values
                }, index=self.index)
            self._rlvi["col_offset"] = \
                self._rlvi["dev"].map(lambda x: self.columns.get_loc(x))
        return(self._rlvi)


    @property
    def clvi(self):
        """
        Determine the last valid index by development period.

        Returns
        -------
        pd.DataFrame
        """
        if self._clvi is None:
            self._clvi = pd.DataFrame({
                "origin": self.apply(lambda v: v.last_valid_index(), axis=0).values
                }, index=self.columns
                )
            self._clvi["row_offset"] = \
                self._clvi["origin"].map(lambda v: self.index.get_loc(v))
        return(self._clvi)


    @property
    def latest(self):
        """
        Return the values on the triangle's latest diagonal. Loss amounts
        are given, along with the associated origin year and development
        period. The latest loss amount by origin year alone can be  obtained
        by calling ``self.latest_by_origin``, or by development period by
        calling by ``self.latest_by_devp``.

        Returns
        -------
        pd.DataFrame
        """
        if self._latest is None:
            lindx = self.apply(lambda devp: devp.last_valid_index(), axis=1)
            dflindx = lindx.to_frame().reset_index(drop=False).rename(
                {0: "dev", "index": self.origin}, axis=1)
            self._latest = dflindx.merge(self.to_tbl(), on=[self.origin, self.dev]).rename(
                {self.value: "latest"}, axis=1)
            self._latest = self._latest[["origin", "dev", "latest"]].sort_index()
        return(self._latest)


    @property
    def latest_by_origin(self):
        """
        Return the latest loss amounts by origin year.

        Returns
        -------
        pd.Series
        """
        if self._latest_by_origin is None:
            self._latest_by_origin = pd.Series(
                data=self.latest["latest"].values, index=self.latest["origin"].values,
                name="latest_by_origin")
        return(self._latest_by_origin.sort_index())


    @property
    def latest_by_devp(self):
        """
        Return the latest loss amounts by development period.

        Returns
        -------
        pd.Series
        """
        if self._latest_by_devp is None:
            self._latest_by_devp = pd.Series(
                data=self.latest["latest"].values, index=self.latest["dev"].values,
                name="latest_by_devp")
        return(self._latest_by_devp.sort_index())


    @property
    def devp(self):
        """
        Return triangle's development periods.

        Returns
        -------
        pd.Series
        """
        if self._devp is None:
            self._devp = pd.Series(self.columns, name="devp")
        return(self._devp.sort_index())


    @property
    def origins(self):
        """
        Return triangle's origin periods.

        Returns
        -------
        pd.Series
        """
        if self._origins is None:
            self._origins = pd.Series(self.index, name="origin")
        return(self._origins.sort_index())


    @property
    def maturity(self):
        """
        Return the maturity for each origin period.

        Returns
        -------
        ps.Series
        """
        if self._maturity is None:
            dfind, matlist = (1 - self.triind), list()
            for i in range(dfind.index.size):
                lossyear = dfind.index[i]
                maxindex = dfind.loc[lossyear].values.nonzero()[0].max()
                itermatur = dfind.columns[maxindex]
                matlist.append(itermatur)
            self._maturity = pd.Series(data=matlist, index=self.index, name="maturity")
        return(self._maturity.sort_index())



    def diagonal(self, offset=0):
        """
        Return triangle values at given offset. When ``offset=0``, returns
        latest diagonal.

        Parameters
        ----------
        offset: int
            Negative integer value (or 0) representing the diagonal to return.
            To return the second diagonal, set ``offset=-1``. If abs(offset)
            exceeds (number of development periods - 1), ``ValueError`` is raised.
            Default value is 0 (represents latest diagonal).

        Returns
        -------
        pd.Series
        """
        if np.abs(offset) > (self.devp.size - 1):
            raise ValueError(
                "abs(offset) cannot exceed the number of development periods."
                )
        df = self.latest.copy()
        df = df.reset_index(drop=False).rename(
            {"index": "origin_indx"}, axis=1)[["origin_indx"]]
        df["dev_indx"] = df["origin_indx"].values[::-1]
        df["dev_indx"] = df["dev_indx"] + offset
        df = df[df.dev_indx >= 0].reset_index(drop=True)
        df = df.assign(
            origin=df["origin_indx"].map(lambda v: self.origins[v]),
            dev=df["dev_indx"].map(lambda v: self.devp[v]),
            value=df.apply(lambda rec: self.iat[rec.origin_indx, rec.dev_indx], axis=1)
            )
        return(df[["origin", "dev", "value"]])



    def to_tbl(self, dropna=True):
        """
        Transform triangle instance into a tabular representation.

        Parameters
        ----------
        dropna: bool
            Should records with NA values be dropped? Default value is True.

        Returns
        -------
        pd.DataFrame
        """
        tri = self.reset_index(drop=False).rename({"index": "origin"}, axis=1)
        df = pd.melt(tri, id_vars=[self.origin], var_name=self.dev, value_name=self.value)
        if dropna:
            df = df[~np.isnan(df[self.value])]
        df = df.astype({self.origin: int, self.dev: int, self.value: float})
        df = df[[self.origin, self.dev, self.value]].sort_values(by=[self.origin, self.dev])
        return(df.reset_index(drop=True))


    def __str__(self):
        formats = {devp: "{:,.0f}".format for devp in self.columns}
        return(self.to_string(formatters=formats))


    def __repr__(self):
        formats = {devp: "{:,.0f}".format for devp in self.columns}
        return(self.to_string(formatters=formats))



class _BaseIncrTriangle(_BaseTriangle):
    """
    Internal incremental triangle class definition.
    """
    def __init__(self, data, origin=None, dev=None, value=None):
        """
        Parameters
        ----------
        data: pd.DataFrame
            The dataset to be transformed into a triangle instance.
            ``data`` must be tabular loss data with at minimum columns
            representing the origin/acident year, development
            period and value of interest, given by ``origin``, ``dev``
            and ``value`` respectively.

        origin: str
            The fieldname in ``data`` representing origin year.

        dev: str
            The fieldname in ``data`` representing development period.

        value: str
            The fieldname in ``data`` representing loss amounts.
        """
        # Replace NaN values with 1.0 in value column.
        # data.loc[np.where(np.isnan(data.value.values))[0], "value"] = 1.
        super().__init__(data, origin=origin, dev=dev, value=value)



class IncrTriangle(_BaseIncrTriangle):
    """
    Public incremental triangle class definition.

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be transformed into a triangle instance.
        ``data`` must be tabular loss data with at minimum columns
        representing the origin/acident year, development
        period and value of interest, given by ``origin``, ``dev``
        and ``value`` respectively.

    origin: str
        The fieldname in ``data`` representing origin year.

    dev: str
        The fieldname in ``data`` representing development period.

    value: str
        The fieldname in ``data`` representing loss amounts.
    """
    def __init__(self, data, origin=None, dev=None, value=None):

        super().__init__(data, origin=origin, dev=dev, value=value)


    def to_cum(self):
        """
        Transform triangle instance into cumulative representation.

        Returns
        -------
        trikit.triangle.CumTriangle
        """
        return(CumTriangle(self.to_tbl(), origin="origin", dev="dev", value="value"))



class _BaseCumTriangle(_BaseTriangle):
    """
    Internal cumulative triangle class definition. Transforms ``data`` into a
    cumulative triangle instance.

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be transformed into a triangle instance.
        ``data`` must be tabular loss data with at minimum columns
        representing the origin/acident year, development
        period and incremental value of interest, given by ``origin``,
        ``dev`` and ``value`` respectively.

    origin: str
        The fieldname in ``data`` representing the origin year.

    dev: str
        The fieldname in ``data`` representing the development period.

    value: str
        The fieldname in ``data`` representing incremental loss amounts.
    """
    def __init__(self, data, origin="origin", dev="dev", value="value"):

        # Replace NaN values with 1.0 in value column.
        data["cumval"] = data.groupby([origin], as_index=False)[value].cumsum()
        data = data.drop(value, axis=1)
        data = data.rename(columns={"cumval": value})
        super().__init__(data=data, origin=origin, dev=dev, value=value)

        # Properties.
        self._a2a_assignment = None
        self._ranked_a2a = None
        self._a2a_lvi = None
        self._weights = None
        self._a2aind = None
        self._a2a = None


    @staticmethod
    def _geometric(vals, weights=None):
        """
        Compute the geometric average of the elements of ``vals``.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Not yet implemented.

        Returns
        -------
        float
        """
        arr = np.asarray(vals, dtype=float)
        return(np.NaN if arr.size == 0 else stats.gmean(arr))


    @staticmethod
    def _simple(vals, weights=None):
        """
        Compute the simple average of elements of ``vals``.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Not yet implemented.

        Returns
        -------
        float
        """
        arr = np.asarray(vals, dtype=float)
        return(np.NaN if arr.size == 0 else arr.mean())


    @staticmethod
    def _medial(vals, weights=None):
        """
        Compute the medial average of elements in ``vals``. Medial average
        eliminates the min and max values, then returns the arithmetic
        average of the remaining items.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Weights to assign specific values in the average computation.
            If None, each value is assigned equal weight.

        Returns
        -------
        float
        """
        if weights is None:
            w = np.ones(len(vals))
        else:
            w = weights
            if len(w) != len(vals):
                raise ValueError("`vals` and `weights` must have same size")

        # Return first element of arr_all if all array elements are the same.
        arr_all = np.sort(np.asarray(vals, dtype=float))
        if np.all(arr_all == arr_all[0]):
            avg = arr_all[0]

        elif arr_all.shape[0] == 1:
            avg = arr_all[0]

        elif arr_all.shape[0] == 2:
            avg = (w * arr_all).sum() / w.sum()

        else:
            medial_indicies = np.where(
                np.logical_and(arr_all != arr_all.min(), arr_all != arr_all.max())
                )
            arr = arr_all[medial_indicies]
            w = w[medial_indicies]

            if arr.shape[0] == 0:
                avg = np.NaN
            else:
                avg = (w * arr).sum() / w.sum()

        return(avg)


    @property
    def a2a(self):
        """
        Compute adjacent proportions, a.k.a. link ratios.

        Returns
        -------
        pd.DataFrame
        """
        if self._a2a is None:
            self._a2a = self.shift(periods=-1, axis=1) / self
            self._a2a = self._a2a.dropna(axis=1, how="all").dropna(axis=0, how="all")
        return(self._a2a.sort_index())


    @property
    def a2aind(self):
        """
        Determine which cells should be included and which to exclude
        when computing age-to-age averages. Cells populated with 1
        are included, cells populated with 0 are excluded.

        Returns
        -------
        pd.DataFrame
        """
        if self._a2aind is None:
            self._a2aind = self.a2a.map(lambda v: 0 if np.isnan(v) else 1)
        return(self._a2aind)


    @a2aind.setter
    def a2aind(self, update_spec):
        """
        Update ``self.a2aind`` in order to down-weight ldfs in chain ladder
        calculation.

        Parameters
        ----------
        update_spec: tuple
            3-tuple consisting of ``(index, column, value)``, representing
            the intersection point of the ``self.a2a`` target cell, and the
            value used to update it. ``value`` must be either 0 or 1.

        Examples
        --------
        Load raa sample dataset, and remove a highly-leveraged age-to-age
        factor from influencing the ldf calculation::

            In [1]: import trikit
            In [2]: raa = trikit.load(dataset="raa")
            In [3]: tri = trikit.totri(data=raa)
            In [4]: tri.a2a.iloc[:, :1]
            Out[1]:
                          1
            1981   1.649840
            1982  40.424528
            1983   2.636950
            1984   2.043324
            1985   8.759158
            1986   4.259749
            1987   7.217235
            1988   5.142117
            1989   1.721992


        To remove the link ratio at origin year 1982 and development
        period 1, run the following::

            In [1]: tri.a2aind = (1982, 1, 0)
            In [2]: tri.a2aind
            Out[1]:
                  1  2  3  4  5  6  7  8  9
            1981  1  1  1  1  1  1  1  1  1
            1982  0  1  1  1  1  1  1  1  0
            1983  1  1  1  1  1  1  1  0  0
            1984  1  1  1  1  1  1  0  0  0
            1985  1  1  1  1  1  0  0  0  0
            1986  1  1  1  1  0  0  0  0  0
            1987  1  1  1  0  0  0  0  0  0
            1988  1  1  0  0  0  0  0  0  0
            1989  1  0  0  0  0  0  0  0  0


        Notice that the value at (1982, 1) is 0. To change it back
        to 1, simply run::

            In [1]: tri.a2aind = (1982, 1, 1)
            In [2]: tri.a2aind
            Out[1]:
                  1  2  3  4  5  6  7  8  9
            1981  1  1  1  1  1  1  1  1  1
            1982  1  1  1  1  1  1  1  1  0
            1983  1  1  1  1  1  1  1  0  0
            1984  1  1  1  1  1  1  0  0  0
            1985  1  1  1  1  1  0  0  0  0
            1986  1  1  1  1  0  0  0  0  0
            1987  1  1  1  0  0  0  0  0  0
            1988  1  1  0  0  0  0  0  0  0
            1989  1  0  0  0  0  0  0  0  0


        Note also that ``self.a2aind`` may be updated using DataFrame
        methods directly::

            In [1]: tri.a2aind.at[1982, 1] = 0
            In [2]: tri.a2aind
            Out[1]:
                  1  2  3  4  5  6  7  8  9
            1981  1  1  1  1  1  1  1  1  1
            1982  0  1  1  1  1  1  1  1  0
            1983  1  1  1  1  1  1  1  0  0
            1984  1  1  1  1  1  1  0  0  0
            1985  1  1  1  1  1  0  0  0  0
            1986  1  1  1  1  0  0  0  0  0
            1987  1  1  1  0  0  0  0  0  0
            1988  1  1  0  0  0  0  0  0  0
            1989  1  0  0  0  0  0  0  0  0
        """
        indx, column, value = update_spec
        self._a2aind.at[indx, column] = value


    @property
    def a2a_lvi(self):
        """
        Reference to last valid index for triangle age-to-age factors.

        Returns
        -------
        pd.DataFrame
        """
        # Bind reference to last valid index by column for self.tri.a2a.
        if self._a2a_lvi is None:
            self._a2a_lvi = pd.DataFrame({
                "origin": self.a2a.apply(lambda v: v.last_valid_index(), axis=0).values
            }, index=self.a2a.columns
            )
            self._a2a_lvi["row_offset"] = self._a2a_lvi["origin"].map(lambda v: self.index.get_loc(v))
        return(self._a2a_lvi)


    @property
    def ranked_a2a(self):
        """
        Construct triangle of ranked age-to-age factors for use in development
        period correlation testing.

        Returns
        -------
        pd.DataFrame
        """
        if self._ranked_a2a is None:
            a2a_lvi = self.a2a_lvi
            rank_list = []
            for devp_indx, devp in enumerate(a2a_lvi.index[:-1]):
                last_valid_origin = a2a_lvi[a2a_lvi.index == devp].origin.item()
                a2a_ii = self.a2a.loc[:, devp]
                r_ii = a2a_ii.rank().to_frame().rename({devp: "r_{}".format(devp_indx + 1)}, axis=1)
                s_ii = a2a_ii[a2a_ii.index < last_valid_origin].rank().to_frame().rename(
                    {devp: "s_{}".format(devp_indx + 2)}, axis=1
                    )
                rank_list.append(r_ii)
                rank_list.append(s_ii)
            self._ranked_a2a = pd.concat(rank_list[1:-1], axis=1).dropna(how="all")
        return(self._ranked_a2a)


    @property
    def a2a_assignment(self):
        """
        Identify triangle age-to-age factors into high and low categories based
        on value relative to the median for a given development period. Factors
        in excess of the median are assigned a value of +1. Age-to-age factors
        with value less than the median are assigned a value of -1. For
        development periods with an odd number of values, the true median is set
        to 0. Returned DataFrame has same dimensionality as self.tri.a2a.

        Returns
        -------
        pd.DataFrame
        """
        if self._a2a_assignment is None:
            self._a2a_assignment = pd.DataFrame(columns=self.a2a.columns, index=self.a2a.index)
            dfmedian = self.a2a.median(axis=0)
            for devp in self.a2a.columns:
                devp_median = dfmedian[devp]
                for r_indx, a2a in enumerate(self.a2a[devp]):
                    origin = self.a2a.index[r_indx]
                    if not np.isnan(a2a):
                        if np.allclose(a2a, devp_median):
                            self._a2a_assignment.at[origin, devp] = 0
                        elif a2a > devp_median:
                            self._a2a_assignment.at[origin, devp] = 1
                        else:
                            self._a2a_assignment.at[origin, devp] = -1
        return(self._a2a_assignment)



    def a2a_avgs(self):
        """
        Compute age-to-age factors based on ``self.a2a`` table of adjacent
        proportions. Averages computed include "simple", "geometric", "medial"
        and "weighted".

        Returns
        -------
        pd.DataFrame
        """
        _nbr_periods = list(range(1, self.a2a.shape[0])) + [0]
        indxstrs = list()

        # Create lookup table for average functions.
        avgfuncs = {"simple": self._simple, "geometric": self._geometric,
                    "medial": self._medial, "weighted": None}

        # Remove `0` entry, and add as last element of list.
        ldf_avg_lst = list(itertools.product(avgfuncs.keys(), _nbr_periods))

        indxstrs = [
            "all-" + str(ii[0]) if ii[1] == 0 else "{}-{}".format(ii[0], ii[1])
                for ii in ldf_avg_lst
            ]

        _a2a_avgs = pd.DataFrame(index=indxstrs, columns=self.a2a.columns)
        a2a_adj = self.a2a * self.a2aind

        for a in enumerate(ldf_avg_lst):
            duration, avgtype, indxpos = a[1][1], a[1][0], a[0]
            indxstr, iterfunc = indxstrs[indxpos], avgfuncs[avgtype]

            for col in range(a2a_adj.shape[1]):
                itercol, colstr = a2a_adj.iloc[:, col], a2a_adj.columns[col]

                if avgtype == 'weighted':

                    t_ic_1, t_ic_2 = self.iloc[:, col], self.iloc[:, (col + 1)]
                    # Find first NaN value in t_ic_2.
                    first_nan_year = t_ic_2.index[t_ic_2.count():][0]
                    first_nan_indx = t_ic_2.index.searchsorted(first_nan_year)
                    final_cell_indx = first_nan_indx
                    if duration == 0:
                        first_cell_indx = 0
                    else:
                        first_cell_indx = (final_cell_indx - duration) if \
                            (final_cell_indx - duration) >= 0 else 0

                    # Divide sum of t_ic_2 by t_ic_1.
                    sum_ic_2 = t_ic_2[first_cell_indx:final_cell_indx].sum()
                    sum_ic_1 = t_ic_1[first_cell_indx:final_cell_indx].sum()

                    try:
                        iteravg = (sum_ic_2 / sum_ic_1)
                    except ZeroDivisionError:
                        iteravg = np.NaN

                else:
                    # Find index of first row with NaN.
                    if any(itercol.map(lambda x: np.isnan(x))):
                        first_nan_year = itercol.index[itercol.apply(lambda x: np.isnan(x))][0]
                        first_nan_indx = itercol.index.searchsorted(first_nan_year)
                        final_cell_indx = first_nan_indx
                        if duration == 0:
                            first_cell_indx = 0
                        else:
                            first_cell_indx = (final_cell_indx - duration) if \
                                              (final_cell_indx - duration) >= 0 else 0

                    else:
                        # itercol has 0 NaN's.
                        final_cell_indx = len(itercol)
                        first_cell_indx = 0 if duration == 0 else (final_cell_indx - duration)
                    try:
                        link_ratios = itercol[first_cell_indx:final_cell_indx]
                        iteravg = iterfunc(link_ratios[link_ratios > 0])
                    except ZeroDivisionError:
                        iteravg = np.NaN
                _a2a_avgs.loc[indxstr, colstr] = iteravg
        return(_a2a_avgs)



class CumTriangle(_BaseCumTriangle):
    """
    Cumulative triangle class definition.
    """
    def __init__(self, data, origin=None, dev=None, value=None):

        super().__init__(data, origin=origin, dev=dev, value=value)


    def to_incr(self):
        """
        Obtain incremental triangle based on cumulative triangle instance.

        Returns
        -------
        trikit.triangle.IncrTriangle

        Examples
        --------
        Convert existing cumulative triangle instance into an instance of
        ``trikit.triangle.IncrTriangle``::

            In [1]: from trikit import load, totri
            In [2]: cumtri = totri(load("raa"))
            In [3]: incrtri = cumtri.to_incr()
            In [4]: type(incrtri)
            Out[4]: triangle.IncrTriangle
        """
        incrtri = pd.DataFrame(self).diff(axis=1)
        incrtri.iloc[:, 0] = self.iloc[:, 0]
        incrtri = incrtri.reset_index(drop=False).rename({"index": "origin"}, axis=1)
        df = pd.melt(incrtri, id_vars=["origin"], var_name="dev", value_name="value")
        df = df[~np.isnan(df["value"])].astype({"origin": int, "dev": int, "value": float})
        df = df.sort_values(by=["origin", "dev"]).reset_index(drop=True)
        return(IncrTriangle(df, origin="origin", dev="dev", value="value"))


    def plot(self, display="combined", **kwargs):
        """
        Plot cumulative loss development over a single set of axes or
        as faceted-by-origin exhibit.

        Parameters
        ----------
        view: {"combined", "faceted"}
            Whether to display cumulative loss development in a single or faceted view.
            Default value is ``"combined"``.

        kwargs: dict

            Options for combined view:

                cmap: str
                    Selected matplotlib color map. For additional options, visit:
                    https://matplotlib.org/tutorials/colors/colormaps.html.

            Options for faceted view:

                color: str
                    Color to plot loss development in each facet. Default value is "#334488".

                axes_style: str
                    Aesthetic style of plots. Defaults to "darkgrid". Other options
                    include: {whitegrid, dark, white, ticks}.

                context: str
                    Set the plotting context parameters. According to the seaborn
                    documentation, This affects things like the size of the labels,
                    lines, and other elements of the plot, but not the overall style.
                    Defaults to ``"notebook"``. Additional options include
                    {"paper", "talk", "poster"}.
        """
        if kwargs is not None:
            kwds = kwargs
        else:
            kwds = {}

        if display.startswith("f"):
            self._faceted_view(**kwds)
        else:
            self._combined_view(**kwds)


    def _combined_view(self, **kwargs):
        """
        Visualize triangle loss development using a combined view.

        Parameters
        ----------
        cmap: str
            Selected matplotlib color map. For additional options, visit:
            https://matplotlib.org/tutorials/colors/colormaps.html.

        kwargs: dict
            Additional plot styling options.
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm

        pltkwargs = dict(
            marker="s", markersize=5, alpha=1, linestyle="-", linewidth=1.5,
            figsize=(9.5, 6.5), cmap="gist_rainbow",
            )

        if kwargs:
            pltkwargs.update(kwargs)

        data = self.to_tbl()
        grps = data.groupby("origin", as_index=False)
        data_list = [grps.get_group(ii) for ii in self.origins]
        xticks = np.sort(data.dev.unique())

        # Get unique hex color for each unique origin period.
        fcolors = cm.get_cmap(pltkwargs["cmap"], len(self.origins))
        colors_rgba = [fcolors(ii) for ii in np.linspace(0, 1, len(self.origins))]
        colors_hex = [mpl.colors.to_hex(ii, keep_alpha=False) for ii in colors_rgba]
        markers = ["o", "v", "^", "s", "p", "D", "d", "h"]

        fig, ax = plt.subplots(1, 1, figsize=pltkwargs["figsize"], tight_layout=True)

        ax.set_title("Loss Development by Origin", fontsize=9, loc="left")

        for ii, hex_color, dforg in zip(range(len(colors_hex)), colors_hex, data_list):
            xx = dforg["dev"].values
            yy = dforg["value"].values
            marker = markers[ii % len(markers)]
            yy_divisor = 1  # 1000 if np.all(yy>1000) else 1
            yy_axis_label = "(000's)" if yy_divisor == 1000 else ""
            ax.plot(
                xx, yy / yy_divisor, color=hex_color,
                linewidth=pltkwargs["linewidth"], linestyle=pltkwargs["linestyle"],
                label=dforg.origin.values[0], marker=marker,
                markersize=pltkwargs["markersize"]
                )

        # Reduce thickness of plot outline.
        for axis in ["top", "bottom", "left", "right"]:
            ax.spines[axis].set_linewidth(0.5)

        ax.get_yaxis().set_major_formatter(mpl.ticker.FuncFormatter(lambda v, p: format(int(v), ",")))
        ax.set_xlabel("dev", fontsize=8)
        ax.set_ylabel(yy_axis_label, fontsize=8)
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0)
        ax.set_xticks(xticks)
        ax.tick_params(axis="x", which="major", direction="in", labelsize=8)
        ax.tick_params(axis="y", which="major", direction="in", labelsize=8)
        ax.xaxis.set_ticks_position("none")
        ax.yaxis.set_ticks_position("none")
        ax.grid(True)
        ax.legend(loc="lower right", fancybox=True, framealpha=1, fontsize="x-small")
        plt.show()


    def _faceted_view(self, color="#334488", axes_style="darkgrid", context="notebook",
                      col_wrap=4, **kwargs):
        """
        Visualize triangle loss development using a faceted view.

        Parameters
        ----------
        color: str
            Color to plot loss development in each facet. Default value is "#334488".

        axes_style: str
            Aesthetic style of plots. Defaults to "darkgrid". Other options
            include: {whitegrid, dark, white, ticks}.

        context: str
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Defaults to ``"notebook"``. Additional options include
            {"paper", "talk", "poster"}.

        kwargs: dict
            Additional plot styling options.
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        pltkwargs = dict(
            color="#334488", axes_style="darkgrid", context="notebook",
            col_wrap=4, marker="s", markersize=5, alpha=1, linestyle="-",
            linewidth=1.5, figsize=(9, 6), cmap="hsv"
            )

        if kwargs:
            pltkwargs.update(kwargs)

        sns.set_context(context)
        data = self.to_tbl()

        with sns.axes_style(axes_style):

            pltkwargs = dict(
                marker="o", markersize=7, alpha=1, markeredgecolor="#000000",
                markeredgewidth=.50, linestyle="--", linewidth=.75,
                fillstyle="full", color=color,
                )

            if kwargs:
                pltkwargs.update(kwargs)

            g = sns.FacetGrid(
                data, col="origin", col_wrap=col_wrap, margin_titles=False,
                despine=True, sharex=True, sharey=True,
                )

            g.map(plt.plot, "dev", "value", **pltkwargs)
            g.set_axis_labels("", "")
            g.set_titles("{col_name}", size=8)
            g.set(xticks=np.sort(data.dev.unique()))
            g.set_xticklabels(np.sort(data.dev.unique()), size=7)

            for ii, _ in enumerate(g.axes):
                ax_ = g.axes[ii]

                ylabelss = [jj.get_text() for jj in list(ax_.get_yticklabels())]
                ylabelsn = [float(jj.replace(u"\u2212", "-")) for jj in ylabelss]
                ylabelsn = [jj for jj in ylabelsn if jj >= 0]
                ylabels = ["{:,.0f}".format(jj) for jj in ylabelsn]
                if (len(ylabels) > 0):
                    ax_.set(yticks=ylabelsn)
                    ax_.set_yticklabels(ylabels, size=7)
                ax_.tick_params(
                    axis='x', which='both', bottom=True, top=False, labelbottom=True
                    )
                ax_.xaxis.set_ticks_position("none")
                ax_.yaxis.set_ticks_position("none")

                # Draw border around each facet.
                for _, spine in ax_.spines.items():
                    spine.set_visible(True)
                    spine.set_color("#000000")
                    spine.set_linewidth(.50)

        plt.show()


    def base_cl(self, sel="all-weighted", tail=1.0):
        """
        Produce chain ladder reserve estimates based on cumulative triangle instance.

        Parameters
        ----------
        sel: str, pd.Series or array_like
            If ``sel`` is a string, the specified loss development patterns will be
            the associated entry from ``self.tri.a2a_avgs``.
            If ``sel`` is array_like, values will be used in place of loss development
            factors computed from the traingle directly. For a triangle with n development
            periods, ``sel`` should be array_like with length n - 1.
            Defaults to "all-weighted".

        tail: float
            Chain ladder tail factor. Defaults to 1.0.

        Examples
        --------
        Generate chain ladder reserve point estimates using the raa dataset.
        ``tri`` is first created using the raa dataset::

            In [1]: import trikit
            In [2]: tri = trikit.load("raa", tri_type="cum")
            In [4]: cl = tri.base_cl()


        Perform standard chain ladder, using non-default values for ``sel`` and ``tail``::

            In [5]: cl = tri.base_cl(sel="medial-5", tail=1.015)


        Passing a custom sequence of loss development factors::

            In [6]: ldfs = [5., 2.5, 1.25, 1.15, 1.10, 1.05, 1.025, 1.01, 1.005,]
            In [7]: cl = tri.base_cl(sel=ldfs, tail=1.001)
        """
        kwds = dict(sel=sel, tail=tail)
        return(BaseChainLadder(self).__call__(**kwds))


    def boot_cl(self, sims=1000, q=[.75, .95], procdist="gamma", parametric=False,
                two_sided=False, interpolation="linear", random_state=None):
        """
        Estimate reserves and the distribution of reserve outcomes by origin and in
        total via bootstrap resampling. The estimated distribution of losses assumes
        development is completen by the final development period in order to avoid the
        complication of modeling a tail factor.

        Parameters
        ----------
        sims: int
            The number of bootstrap simulations to perform. Defaults to 1000.

        q: array_like of float or float
            Quantile or sequence of quantiles to compute, which must be
            between 0 and 1 inclusive.

        procdist: str
            The distribution used to incorporate process variance. Currently,
            this can only be set to "gamma".

        two_sided: bool
            Whether to include the two_sided interval in summary output. For example,
            if ``two_sided==True`` and ``q=.95``, the 2.5th and 97.5th quantiles of the
            bootstrapped reserve  distribution will be returned [(1 - .95) / 2, (1 + .95) / 2].
            When False, only the specified quantile(s) will be computed. Defaults
            to False.

        parametric: bool
            If True, fit standardized residuals to a normal distribution via maximum likelihood,
            and sample from the resulting distribution. Otherwise, values are sampled with
            replacement from the collection of standardized residuals. Defaults to False.

        interpolation: {"linear", "lower", "higher", "midpoint", "nearest"}
            Optional parameter which specifies the interpolation method to use
            when the desired quantile lies between two data points i < j. See
            ``numpy.quantile`` for more information. Default value is "linear".

        random_state: np.random.RandomState
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by np.random.

        Returns
        -------
        BootstrapChainLadderResult


        Examples
        --------
        Generate boostrap chain ladder reserve estimates. ``tri`` is first created
        using the raa dataset::

            In [1]: import trikit
            In [2]: tri = trikit.load("raa", tri_type="cum")
            In [3]: bcl = tri.boot_cl()
        """
        kwds = dict(
            sims=sims, q=q, procdist=procdist, parametric=parametric, two_sided=two_sided,
            interpolation=interpolation, random_state=random_state
            )
        return(BootstrapChainLadder(self).__call__(**kwds))


    def mack_cl(self, alpha=1, tail=1.0, dist="lognorm", q=[.75, .95], two_sided=False):
        """
        Return a summary of ultimate and reserve estimates resulting from
        the application of the development technique over self.tri. Summary
        DataFrame is comprised of origin year, maturity of origin year, loss
        amount at latest evaluation, cumulative loss development factors,
        projected ultimates and the reserve estimate, by origin year and in
        aggregate.

        ### TODO ###
        Allow for tail factor other than 1.0.

        Parameters
        ----------
        alpha: {0, 1, 2}
            * 0: Straight average of observed individual link ratios.
            * 1: Historical Chain Ladder age-to-age factors.
            * 2: Regression of :math:`C_{k+1}` on :math:`C_{k}` with 0 intercept.

        tail: float
            Tail factor. Currently not implemented. Will be available in
            a future release.

        dist: {"norm", "lognorm"}
            The distribution function chosen to approximate the true
            distribution of reserves by origin period. In Mack[1], if the
            volume of outstanding claims is large enough, due to the central
            limit theorem, we can assume that the distribution function is
            Normal with expected value equal to the point estimate given by
            :math:`R_{i}` and standard deviation equal to the standard error of
            :math:`R_{i}`, :math:`s.e.(R_{i})`. It is also noted that if the true
            distribution of reserves is skewed, the Normal may not serve as a
            good approximation, and it may be preferrable to opt for
            the Log-normal distribution.

            * If ``dist="norm"``, the Normal distribution will be used to
            estimate reserve quantiles.

            * If ``dist="lognorm"``, the Log-normal distribution will be used
            to estimate reserve quantiles.

        q: array_like of float
            Quantile or sequence of quantiles to compute, which must be
            between 0 and 1 inclusive.

        two_sided: bool
            Whether the two_sided interval should be included in summary
            output. For example, if ``two_sided==True`` and ``q=.95``, then
            the 2.5th and 97.5th quantiles of the estimated reserve
            distribution will be returned [(1 - .95) / 2, (1 + .95) / 2]. When
            False, only the specified quantile(s) will be computed. Defaults
            to False.

        Returns
        -------
        MackChainLadderResult

        Examples
        --------
        Generate Mack chain ladder reserve estimates. ``tri`` is first created
        using the raa dataset. In the call to ``mack_cl``, ``alpha`` is set to
        2, and ``two_sided=True``::

            In [1]: import trikit
            In [2]: tri = trikit.load("raa", tri_type="cum")
            In [3]: mcl = tri.mack_cl(alpha=2, two_sided=True)
        """
        kwds = dict(alpha=alpha, tail=tail, dist=dist, q=q, two_sided=two_sided)
        return(MackChainLadder(self).__call__(**kwds))


    def glm(self, var_power=1):
        """
        Generate reserve estimates via Generalized Linear Model framework.
        Note that ``glm_estimator`` assumes development is complete by the final
        development period. GLMs are fit using statsmodels Tweedie family
        with log link.
        """
        raise NotImplementedError("glm reserve estimator not yet implemented.")



def totri(data, tri_type="cum", data_format="incr", data_shape="tabular",
          origin="origin", dev="dev", value="value"):
    """
    Create a triangle object based on ``data``. ``tri_type`` can be one of
    "incr" or "cum", determining whether the resulting triangle represents
    incremental or cumulative losses/counts.
    If ``data_shape="triangle"``, ``data`` is assumed to be structured as a
    runoff triangle, indexed by origin with columns representing development
    periods. If ``data_shape="tabular"``, data is assumed to be tabular with at
    minimum columns ``origin``, ``dev`` and ``value``, which represent origin
    year, development period and metric of interest respectively.
    ``data_format`` specifies whether the metric of interest are cumulative
    or incremental in nature. Default value is "incr".

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be coerced into a triangle instance. ``data`` can be
        tabular loss data, or a dataset (pandas DataFrame) formatted as a
        triangle, but not typed as such. In the latter case,
        ``data_shape`` should be set to "triangle".

    tri_type: {"cum", "incr"}
        Either "cum" or "incr". Specifies how the measure of interest (losses,
        counts, alae, etc.) should be represented in the returned triangle
        instance.

    data_format: {"cum", "incr"}
        Specifies the representation of the metric of interest in ``data``.
        Default value is "incr".

    data_shape:{"tabular", "triangle"}
        Indicates whether ``data`` is formatted as a triangle instead of
        tabular loss data. In some workflows, triangles may have already
        been created and are available. In such cases, the triangle-formatted
        data is read into a DataFrame, then coerced into the desired triangle
        representation. Default value is "tabular".

    origin: str
        The field in ``data`` representing origin year. When ``data_shape="triangle"``,
        ``origin`` is ignored. Default value is "origin".

    dev: str
        The field in ``data`` representing development period. When
        ``data_shape="triangle"``,  ``dev`` is ignored. Default value is
        "dev".

    value: str
        The field in ``data`` representing the metric of interest (losses, counts, etc.).
        When ``data_shape="triangle"``, ``value`` is ignored. Default value is "value".

    Returns
    -------
    {trikit.triangle.IncrTriangle, trikit.triangle.CumTriangle}

    Examples
    --------
    Create incremental triangle based on RAA dataset::

        In [1]: from trikit import load, totri
        In [2]: df = load("raa")
        In [3]: tri = totri(df, tri_type="incr")

    """
    if data_shape == "triangle":

        if data_format.lower().strip().startswith("i"):
            # data is in incremental triangle format (but not typed as such).
            incrtri = data.reset_index(drop=False).rename({"index": "origin"}, axis=1)
            df = pd.melt(incrtri, id_vars=["origin"], var_name="dev", value_name="value")

        elif data_format.lower().strip().startswith("c"):
            # data is in cumulative triangle format (but not typed as such).
            incrtri = data.diff(axis=1)
            incrtri.iloc[:, 0] = data.iloc[:, 0]
            incrtri = incrtri.reset_index(drop=False).rename({"index": "origin"}, axis=1)
            df = pd.melt(incrtri, id_vars=["origin"], var_name="dev", value_name="value")
            df = df[~pd.isnull(df["value"])].astype({"origin": int, "dev": int, "value": float})

        else:
            raise NameError("Invalid data_format argument: `{}`.".format(tri_type))

        df = df[~pd.isnull(df["value"])].astype({"origin": int, "dev": int, "value": float})
        df = df.sort_values(by=["origin", "dev"]).reset_index(drop=True)

    elif data_shape == "tabular":

        if data_format.lower().strip().startswith("c"):
            df = data.rename({value: "cum"}, axis=1)
            df["incr"] = df.groupby([origin])["cum"].diff(periods=1)
            df["incr"] = np.where(np.isnan(df["incr"]), df["cum"], df["incr"])
            df = df.drop("cum", axis=1).rename({"incr": value}, axis=1)
        else:
            df = data
    else:
        raise NameError("Invalid data_shape argument: `{}`.".format(data_shape))

    df = df.reset_index(drop=True)

    # Transform df to triangle instance.
    if tri_type.lower().startswith("i"):
        tri = IncrTriangle(data=df, origin=origin, dev=dev, value=value)

    elif tri_type.lower().startswith("c"):
        tri = CumTriangle(data=df, origin=origin, dev=dev, value=value)

        # Replace missing actuals.
        for origin_ in tri.index:
            origin_indx = tri.index.get_loc(origin_)
            origin_init_val = tri.iat[origin_indx, 0]
            if np.isnan(origin_init_val):
                tri.iat[origin_indx, 0] = 1.

            for devp_indx, devp_ in enumerate(tri.columns[1:], start=1):
                triind_val = tri.triind.iat[origin_indx, devp_indx]
                if triind_val == 0:
                    if np.isnan(tri.iat[origin_indx, devp_indx]):
                        tri.iat[origin_indx, devp_indx] = tri.iat[origin_indx, (devp_indx - 1)]
    return(tri)



"""
BootstrapChainLadder implementation.
"""
import functools
import warnings
import numpy as np
import pandas as pd
from numpy.random import RandomState
from scipy import stats
from .base import BaseRangeEstimator, BaseRangeEstimatorResult


class BootstrapChainLadder(BaseRangeEstimator):
    """
    The purpose of the bootstrap technique is to estimate the predicition
    error of the total reserve estimate and to approximate the predictive
    distribution. It is often impractical to obtain the prediction error
    using an analytical approach due to the complexity of reserve estimators.

    Predicition error is comprised of two components: process error
    and estimation error (Prediction Error = Estimation Error + Process Error).
    The estimation error (parameter error) represents the uncertainty in the
    parameter estimates given that the model is correctly specified. The
    process error is analogous to the variance of a random variable,
    representing the uncertainty in future outcomes.

    The procedure used to generate the predicitive distribution of reserve
    estimates is based on Leong et al. Appendix A, assuming the starting point
    is a triangle of cumulative losses:

    1.  Calculate the all-year volume-weighted age-to-age factors.

    2.  Estimate the fitted historical cumulative paid loss and ALAE
        using the latest diagonal of the original triangle and the
        age-to-age factors from [1] to un-develop the losses.

    3.  Calculate the unscaled Pearson residuals, degrees of freedom
        and scale parameter.

    4.  Calculate the adjusted Pearson residuals.

    5.  Sample with replacement from the adjusted Pearson residuals.

    6.  Calculate the triangle of sampled incremental losses
        (I^ = m + r_adj * sqrt(m)), where I^ = Resampled incremental loss,
        m = Incremental fitted loss (from [2]) and r_adj = Adjusted Pearson
        residuals.

    7.  Using the triangle from [6], project future losses using the
        Chain Ladder method.

    8.  Include Process variance by simulating each incremental future
        loss from a Gamma distribution with mean = I^ and
        variance = I^ * scale parameter.

    9.  Estimate unpaid losses using the Chain Ladder technique.

    10. Repeat for the number of cycles specified.

    The collection of projected ultimates for each origin year over all
    bootstrap cycles comprises the predictive distribtuion of reserve
    estimates.
    Note that the estimate of the distribution of losses assumes
    development is complete by the final development period. This is
    to avoid the complication associated with modeling a tail factor.


    References
    ----------
    1. England, P., and R. Verrall, (2002), *Stochastic Claims Reserving in General
       Insurance*, British Actuarial Journal 8(3): 443-518.

    2. CAS Working Party on Quantifying Variability in Reserve Estimates,
       *The Analysis and Estimation of Loss & ALAE Variability: A Summary Report*,
       Casualty Actuarial Society Forum, Fall 2005.

    3. Leong et al., (2012), *Back-Testing the ODP Bootstrap of the Paid
       Chain-Ladder Model with Actual Historical Claims Data*, Casualty Actuarial
       Society E-Forum.

    4. Kirschner, et al., *Two Approaches to Calculating Correlated Reserve
       Indications Across Multiple Lines of Business* Appendix III, Variance
       Journal, Volume 2/Issue 1.

    5. Shapland, Mark R., (2016), *Using the ODP Bootstrap Model: A
       Practicioner's Guide*, CAS Monograph Series Number 4: Casualty Actuarial
       Society, 2016.
    """
    def __init__(self, cumtri):
        """
        The BootstrapChainLadder class definition.

        Parameters
        ----------
        cumtri: triangle.CumTriangle
            A cumulative triangle instance.
        """
        super().__init__(cumtri=cumtri)

        self._dfrlvi = None
        self._dof = None


    def __call__(self, sims=1000, q=[.75, .95], procdist="gamma", parametric=False,
                 two_sided=False, interpolation="linear", random_state=None):
        """
        ``BootstrapChainLadder`` simulation initializer. Generates predictive
        distribution of reserve outcomes by origin and in total.

        The estimated distribution of losses assumes development is complete
        by the final development period in order to avoid the complication of
        modeling a tail factor.

        Parameters
        ----------
        sims: int
            The number of bootstrap simulations to perform. Defaults to 1000.

        q: array_like of float or float
            Quantile or sequence of quantiles to compute, which must be
            between 0 and 1 inclusive.

        procdist: str
            The distribution used to incorporate process variance. Currently,
            this can only be set to "gamma".

        two_sided: bool
            Whether the two_sided interval should be included in summary
            output. For example, if ``two_sided==True`` and ``q=.95``, then
            the 2.5th and 97.5th quantiles of the bootstrapped reserve
            distribution will be returned [(1 - .95) / 2, (1 + .95) / 2]. When
            False, only the specified quantile(s) will be computed. Defaults
            to False.

        parametric: bool
            If True, fit standardized residuals to a normal distribution, and
            sample from this parameterized distribution. Otherwise, bootstrap
            procedure samples with replacement from the collection of
            standardized residuals. Defaults to False.

        interpolation: {"linear", "lower", "higher", "midpoint", "nearest"}
            This optional parameter specifies the interpolation method to use
            when the desired quantile lies between two data points i < j. See
            ``numpy.quantile`` for more information. Default value is "linear".

        random_state: np.random.RandomState
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by np.random.

        Returns
        -------
        BootstrapChainLadderResult
        """
        ldfs = self._ldfs(sel="all-weighted")
        cldfs = self._cldfs(ldfs=ldfs)
        maturity = self.tri.maturity.astype(str)
        latest = self.tri.latest_by_origin
        trisqrd = self._trisqrd(ldfs=ldfs)

        # Obtain reference to BootstrapChainLadder estimates.
        tri_fit_cum = self._tri_fit_cum(ldfs=ldfs)
        tri_fit_incr = self._tri_fit_incr(fitted_tri_cum=tri_fit_cum)
        unscld_residuals = self._resid_us(fitted_tri_incr=tri_fit_incr)
        adjust_residuals = self._resid_adj(resid_us=unscld_residuals)
        scale_param = self._scale_param(resid_us=unscld_residuals)
        sampling_dist = self._sampling_dist(resid_adj=adjust_residuals)
        dfsamples = self._bs_samples(
            sampling_dist=sampling_dist, fitted_tri_incr=tri_fit_incr,
            sims=sims, parametric=parametric,
            random_state=random_state
            )
        dfldfs = self._bs_ldfs(dfsamples=dfsamples)
        dfforecasts = self._bs_forecasts(dfsamples, dfldfs, scale_param)
        dfprocerror = self._bs_process_error(
            dfforecasts=dfforecasts, scale_param=scale_param, procdist=procdist,
            random_state=random_state
            )

        dfreserves = self._bs_reserves(dfprocerror=dfprocerror)
        ultimates = dfreserves.groupby(["origin"])["ultimate"].mean()
        ultimates[latest.index.min()] = latest[latest.index.min()]
        reserves = pd.Series(ultimates - latest, name="reserve")
        std_error = self._bs_std_error(dfreserves)
        cv = pd.Series(std_error / reserves, name="cv")
        qtls, qtlhdrs = self._qtls_formatter(q=q, two_sided=two_sided)

        # Compile Chain Ladder point estimate summary.
        dfmatur = maturity.to_frame().reset_index(drop=False).rename({"index": "origin"}, axis=1)
        dfcldfs = cldfs.to_frame().reset_index(drop=False).rename({"index": "maturity"}, axis=1)
        dfcldfs["maturity"] = dfcldfs["maturity"].astype(str)
        dfcldfs["emergence"] = 1 / dfcldfs["cldf"]
        dfsumm = dfmatur.merge(dfcldfs, on=["maturity"], how="left").set_index("origin")
        dfsumm.index.name = None
        dflatest = latest.to_frame().rename({"latest_by_origin": "latest"}, axis=1)
        dfsumm = functools.reduce(
            lambda df1, df2: df1.join(df2),
            (dflatest, ultimates.to_frame(), reserves.to_frame(), std_error.to_frame(), cv.to_frame()),
            dfsumm
            )

        # Add "Total" index and set to NaN fields that shouldn't be aggregated.
        dfsumm.loc["total"] = dfsumm.sum()
        dfsumm.loc["total", "maturity"] = ""
        dfsumm.loc["total", ["cldf", "emergence"]] = np.NaN
        dfsumm.loc["total", "std_error"] = std_error["total"]
        dfsumm.loc["total", "cv"] = std_error["total"] / dfsumm.loc["total", "reserve"]

        # Attach quantiles.
        dftotal_res = dfreserves.groupby(["sim"], as_index=False).sum()
        dftotal_res["origin"] = "total"
        dfreserves = pd.concat([dfreserves, dftotal_res])

        for ii, jj in zip(qtls, qtlhdrs):
            dfsumm[jj] = dfsumm.index.map(
                lambda v: np.percentile(
                    dfreserves[dfreserves.origin == v]["reserve"].values,
                    100 * ii, method=interpolation
                    )
                )

        bcl_result = BootstrapChainLadderResult(
            summary=dfsumm, tri=self.tri, ldfs=ldfs, tail=1.0, trisqrd=trisqrd,
            reserve_dist=dfreserves, sims_data=dfprocerror, scale_param=scale_param,
            dof=self.dof, unscaled_residuals=unscld_residuals,
            adjusted_residuals=adjust_residuals,
            sampling_dist=None if parametric else sampling_dist,
            fitted_tri_cum=tri_fit_cum, fitted_tri_incr=tri_fit_incr, sims=sims,
            procdist=procdist, parametric=parametric, q=q, interpolation=interpolation
            )

        return(bcl_result)


    @property
    def dfrlvi(self):
        """
        Transform triangle's last valid origin index into DataFrame format.

        Returns
        -------
        pd.DataFrame
        """
        if self._dfrlvi is None:
            df = self.tri.rlvi.reset_index(drop=False)
            df = df.rename({"index": "origin", "dev": "l_act_dev"}, axis=1)
            self._dfrlvi = df.drop("col_offset", axis=1)
        return(self._dfrlvi)


    def _get_dfcombined(self, dfsamples, dfldfs):
        """
        Merge output of ``self._bs_samples`` and ``self._bs_ldfs``.

        Parameters
        ----------
        dfsamples: pd.DataFrame
            Output from ``self._bs_samples``.

        dfldfs: pd.DataFrame
            Output from ``self._bs_ldfs``.

        Returns
        -------
        pd.DataFrame
        """
        dfcombined = dfsamples.merge(dfldfs, on=["sim", "dev"], how="left")
        dfcombined = dfcombined.merge(self.dfrlvi, on=["origin"], how="left")
        return(dfcombined.reset_index(drop=True).sort_values(by=["sim", "origin", "dev"]))


    @property
    def dof(self):
        """
        Return the degress of freedom.

        Returns
        -------
        int
        """
        if self._dof is None:
            self._dof = self.tri.nbr_cells - (self.tri.columns.size - 1) + self.tri.index.size
        return(self._dof)


    def _scale_param(self, resid_us):
        """
        Return the scale parameter, which is the sum of the squared unscaled
        Pearson residuals over the degrees of freedom. This method is intended
        for internal use only.

        Parameters
        ----------
        resid_us: pd.DataFrame
            Unscaled Pearson residuals, typically output by
            ``self._resid_us``.

        Returns
        -------
        float
        """
        return((resid_us**2).sum().sum() / self.dof)


    def _tri_fit_cum(self, ldfs):
        """
        Return the cumulative fitted triangle using backwards recursion,
        starting with the observed cumulative paid/incurred-to-date along the
        latest diagonal.

        Parameters
        ----------
        ldfs: pd.Series
            Selected ldfs, typically the output of calling ``self._ldfs``.

        Returns
        -------
        pd.DataFrame
        """
        fitted_tri_cum = self.tri.copy(deep=True)

        for ii in range(fitted_tri_cum.shape[0]):
            iterrow = fitted_tri_cum.iloc[ii, :]
            if iterrow.isnull().any():
                # Find first NaN element in iterrow.
                nan_hdr = iterrow.isnull()[iterrow.isnull() == True].index[0]
                nan_idx = fitted_tri_cum.columns.tolist().index(nan_hdr)
                init_idx = nan_idx - 1
            else:
                # If here, iterrow is the most mature exposure period.
                init_idx = fitted_tri_cum.shape[1] - 1

            # Set to NaN any development periods earlier than init_idx.
            fitted_tri_cum.iloc[ii, :init_idx] = np.NaN

            # Iterate over rows, undeveloping triangle from latest diagonal.
            for jj in range(fitted_tri_cum.iloc[ii, :init_idx].size, 0, -1):
                prev_col_idx, curr_col_idx, curr_ldf_idx = jj, jj - 1, jj - 1
                prev_col_val = fitted_tri_cum.iloc[ii, prev_col_idx]
                curr_ldf_val = ldfs.iloc[curr_ldf_idx]
                fitted_tri_cum.iloc[ii, curr_col_idx] = (prev_col_val / curr_ldf_val)
        return(fitted_tri_cum)


    @staticmethod
    def _tri_fit_incr(fitted_tri_cum):
        """
        Return a fitted incremental triangle.

        Parameters
        ----------
        fitted_tri_cum: pd.DataFrame
            Typically the output from ``self._tri_fit_cum``.

        Returns
        -------
        pd.DataFrame
        """
        tri = fitted_tri_cum.diff(axis=1)
        tri.iloc[:, 0] = fitted_tri_cum.iloc[:, 0]
        return(tri)


    def _resid_us(self, fitted_tri_incr):
        """
        Return unscaled Pearson residuals, given by
        :math:`r_{us} = \\frac{I - m}{\\sqrt{|m|}}`, where :math:`r_{us}` represents the
        unscaled Pearson residuals, :math:`I` the actual incremental losses and :math:`m`
        fitted incremental losses.

        Parameters
        ----------
        fitted_tri_incr: pd.DataFrame
            Typically the output from ``self._tri_fit_incr``.

        Returns
        -------
        pd.DataFrame
        """
        # I represents actual incremental losses, m fitted incremental losses.
        I = pd.DataFrame(self.tri.to_incr())
        m = fitted_tri_incr
        return((I - m) / np.sqrt(m.abs()))


    def _resid_adj(self, resid_us):
        """
        Compute and return the adjusted Pearson residuals, given by
        :math:`r_{adj} = \\sqrt{\\frac{N}{dof}} * r_{us}`, where *r_adj*
        represents the adjusted Pearson residuals, *N* the number of triangle cells,
        *dof* the degress of freedom and *r_us* the unscaled Pearson residuals.

        Parameters
        ----------
        resid_us: pd.DataFrame
            Unscaled Pearson residuals, typically output by ``self._resid_us``.

        Returns
        -------
        pd.DataFrame
        """
        return(np.sqrt(self.tri.nbr_cells / self.dof) * resid_us)


    @staticmethod
    def _sampling_dist(resid_adj):
        """
        Return ``resid_adj`` as a 1-dimensional array, which will be sampled
        from with replacement in order to produce synthetic triangles for
        bootstrapping. Any NaN's and 0's present in ``resid_adj`` will not be
        present in the returned array.

        Parameters
        ----------
        resid_adj: pd.DataFrame
            Adjusted Pearson residuals, typically output by ``self._resid_adj``.

        Returns
        -------
        np.ndarray
        """
        resid_ = resid_adj.iloc[:-1, :-1].values.ravel()
        return(resid_[np.logical_and(~np.isnan(resid_), resid_ != 0)])


    def _bs_samples(self, sampling_dist, fitted_tri_incr, sims=1000, parametric=False,
                    random_state=None):
        """
        Return DataFrame containing sims resampled-with-replacement
        incremental loss triangles if ``parametric=False``, otherwise
        random variates from a normal distribution with mean zero and
        variance derived from ``resid_adj``. Randomly generated incremental
        data gets cumulated in preparation for ldf calculation in next
        step.

        Parameters
        ----------
        sampling_dist: np.ndarray
            The residuals from the fitted incremental triangle coerced
            into a one-dimensional numpy array.

        fitted_tri_incr: pd.DataFrame
            The incremental triangle fitted using backwards recursion.
            Typically the output of ``self._tri_fit_incr``.

        sims: int
            The number of bootstrap simulations to run. Defaults to 1000.

        parametric: bool
            If True, fit standardized residuals to a normal distribution, and
            sample from the parameterized distribution. Otherwise, bootstrap
            procedure proceeds by sampling with replacement from the array
            of standardized residuals. Defaults to False.

        random_state: np.random.RandomState
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by np.random.

        Returns
        -------
        pd.DataFrame
        """
        if random_state is not None:
            if isinstance(random_state, int):
                prng = RandomState(random_state)
            elif isinstance(random_state, RandomState):
                prng = random_state
        else:
            prng = RandomState()

        sampling_dist = sampling_dist.flatten()
        fti = fitted_tri_incr.reset_index(drop=False).rename({"index": "origin"}, axis=1)
        dfm = pd.melt(fti, id_vars=["origin"], var_name="dev", value_name="value")
        dfm = dfm[~np.isnan(dfm["value"])].astype({"origin": int, "dev": int, "value": float})

        # Make positive any first development period negative values.
        min_devp = dfm["dev"].min()
        dfm["value"] = np.where(
            np.logical_and(dfm["dev"].values == min_devp, dfm["value"].values < 0),
            1., dfm["value"].values
            )

        dfi = self.tri.to_tbl(dropna=False).drop("value", axis=1)
        dfp = dfi.merge(dfm, how="outer", on=["origin", "dev"])
        dfp["rectype"] = np.where(np.isnan(dfp["value"].values), "forecast", "actual")
        dfp = dfp.rename({"value": "incr"}, axis=1)
        dfp["incr_sqrt"] = np.sqrt(dfp["incr"].values)
        dfrtypes = {"origin": int, "dev": int, "incr": float,
                   "incr_sqrt": float, "rectype": str}
        dfrcols = ["origin", "dev", "incr", "rectype", "incr_sqrt"]

        # Replicate dfp sims times then redefine datatypes.
        dfr = pd.DataFrame(np.tile(dfp, (sims, 1)), columns=dfrcols).astype(dfrtypes)

        # Assign simulation identifier to each record in dfr.
        dfr["sim"] = np.divmod(dfr.index, self.tri.shape[0] * self.tri.shape[1])[0]
        sample_size = dfr.shape[0]

        if parametric:
            # Sample random standard normal residuals.
            dfr["resid"] = prng.normal(loc=0, scale=sampling_dist.std(ddof=1), size=sample_size)
        else:
            # Randomly sample residuals from sampling_dist.
            dfr["resid"] = prng.choice(sampling_dist, sample_size, replace=True)

        # Calcuate resampled incremental and cumulative losses.
        dfr["resid"] = np.where(dfr["rectype"].values == "forecast", np.NaN, dfr["resid"].values)
        dfr = dfr.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True)
        dfr["samp_incr"] = dfr["incr"].values + dfr["resid"].values * dfr["incr_sqrt"].values
        dfr["samp_cum"] = dfr.groupby(["sim", "origin"], as_index=False)["samp_incr"].cumsum()
        return(dfr.reset_index(drop=True))


    def _bs_ldfs(self, dfsamples):
        """
        Compute and return loss development factors for each set of synthetic
        loss data.

        Parameters
        ----------
        dfsamples: pd.DataFrame
            Output from ``self._bs_samples``.

        Returns
        -------
        pd.DataFrame
        """
        keepcols = ["sim", "origin", "dev", "samp_cum", "last_origin"]
        new_col_names = {"index": "dev", "origin": "last_origin", "row_offset": "origin_offset"}
        dflvi = self.tri.clvi.reset_index(drop=False).rename(new_col_names, axis=1)
        dfinit = dfsamples.merge(dflvi, how="left", on=["dev"])
        dfinit = dfinit[keepcols].sort_values(by=["sim", "dev", "origin"])
        df = dfinit[~np.isnan(dfinit["samp_cum"])].reset_index(drop=True)
        df["_aggdev2"] = np.where(df["origin"].values == df["last_origin"].values, 0, df["samp_cum"].values)
        df2 = df.groupby(["sim", "dev"], as_index=False)[["samp_cum", "_aggdev2"]].sum().rename(
            {"samp_cum": "_aggdev1"}, axis=1)
        df2["_aggdev2"] = df2["_aggdev2"].shift(periods=1)
        df2["dev"] = df2["dev"].shift(periods=1)
        dfldfs = df2[df2["_aggdev2"] != 0].dropna(how="any")
        dfldfs["dev"] = dfldfs["dev"].astype(int)
        dfldfs["ldf"] = dfldfs["_aggdev1"] / dfldfs["_aggdev2"]
        return(dfldfs[["sim", "dev", "ldf"]].reset_index(drop=True))


    def _bs_forecasts(self, dfsamples, dfldfs, scale_param):
        """
        Populate lower-right of each simulated triangle using values from
        ``self._bs_samples`` and development factors from ``self._bs_ldfs``.

        Parameters
        ----------
        Parameters
        ----------
        dfsamples: pd.DataFrame
            Output from ``self._bs_samples``.

        dfldfs: pd.DataFrame
            Output from ``self._bs_ldfs``.

        scale_param: float
            the sum of the squared unscaled Pearson residuals over the
            degrees of freedom. Output from ``self._scale_param``.

        Returns
        -------
        pd.DataFrame
        """
        dfcombined = self._get_dfcombined(dfsamples, dfldfs)
        min_origin_year = dfcombined["origin"].min()
        dfcombined["_l_init_indx"] = np.where(
            dfcombined["dev"].values >= dfcombined["l_act_dev"].values,
            dfcombined.index.values, -1
            )
        dfacts = dfcombined[(dfcombined["origin"].values == min_origin_year) |
                            (dfcombined["_l_init_indx"].values == -1)]

        dffcst = dfcombined[~dfcombined.index.isin(dfacts.index)].sort_values(
            by=["sim", "origin", "dev"])
        dffcst["_l_act_indx"] = dffcst.groupby(["sim", "origin"])["_l_init_indx"].transform("min")
        l_act_cum = dffcst.loc[dffcst["_l_act_indx"], "samp_cum"].values
        dffcst["l_act_cum"] = l_act_cum
        dffcst["_cum_ldf"] = dffcst.groupby(["sim", "origin"])["ldf"].transform(
            "cumprod").shift(periods=1)
        dffcst["_samp_cum2"] = dffcst["l_act_cum"].values * dffcst["_cum_ldf"].values
        dffcst["_samp_cum2"] = np.where(
            np.isnan(dffcst["_samp_cum2"].values), 0, dffcst["_samp_cum2"].values
            )
        dffcst["cum_final"] = np.where(
            np.isnan(dffcst["samp_cum"].values), 0,
            dffcst["samp_cum"].values) + dffcst["_samp_cum2"].values

        # Combine forecasts with actuals then compute incremental losses by sim and origin.
        dffcst = dffcst.drop(labels=["samp_cum", "samp_incr"], axis=1).rename(
            columns={"cum_final": "samp_cum"})
        dfsqrd = pd.concat([dffcst, dfacts], sort=True).sort_values(
            by=["sim", "origin", "dev"])
        dfsqrd["_dev1_ind"] = (dfsqrd["dev"].values == 1) * 1
        dfsqrd["_incr_dev1"] = dfsqrd["_dev1_ind"].values * dfsqrd["samp_cum"].values
        dfsqrd["_incr_dev2"] = dfsqrd.groupby(["sim", "origin"])["samp_cum"].diff(periods=1)
        dfsqrd["_incr_dev2"] = np.where(
            np.isnan(dfsqrd["_incr_dev2"].values), 0, dfsqrd["_incr_dev2"].values
            )
        dfsqrd["samp_incr"] = dfsqrd["_incr_dev1"].values + dfsqrd["_incr_dev2"].values
        dfsqrd["var"] = np.abs(dfsqrd["samp_incr"].values * scale_param)
        dfsqrd["sign"] = np.where(dfsqrd["samp_incr"].values > 0, 1, -1)
        dfsqrd = dfsqrd.drop(
            labels=[ii for ii in dfsqrd.columns if ii.startswith("_")], axis=1)
        return(dfsqrd.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))


    @staticmethod
    def _bs_process_error(dfforecasts, scale_param, procdist="gamma", random_state=None):
        """
        Incorporate process error by simulating each incremental future
        loss from ``procdist``. The mean is set to the forecast incremental
        loss amount and variance to `mean x self.scale_param`.
        The parameters for ``procdist`` must be positive. Since the mean
        and variance used to parameterize ``procdist`` depend on the
        resampled incremental losses, it is necessary to incorporate logic
        to address the possibility of negative incremental losses arising
        in the resampling stage. The approach used to handle negative
        incremental values is described in  Shapland[1], and replaces the
        distribution mean with the absolute value of the mean, and the
        variance with the absolute value of the mean multiplied by ``scale_param``.

        Parameters
        ----------
        dfforecasts: pd.DataFrame
            DateFrame of bootstraps forecasts generated within
            ``self._bs_forecasts``.

        scale_param: float
            the sum of the squared unscaled Pearson residuals over the
            degrees of freedom. Available in ``self._scale_param``.

        procdist: str
            Specifies the distribution used to incorporate process error.
            Currently, can only be set to "gamma". Any other distribution
            will result in an error.

        random_state: np.random.RandomState
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by np.random.

        Returns
        -------
        pd.DataFrame
        """
        # Initialize pseudo random number generator.
        if random_state is not None:
            if isinstance(random_state, int):
                prng = RandomState(random_state)
            elif isinstance(random_state, RandomState):
                prng = random_state
        else:
            prng = RandomState()

        # Parameterize distribution for the incorporation of process variance.
        if procdist.strip().lower() == "gamma":
            dfforecasts["param2"] = scale_param
            dfforecasts["param1"] = np.abs(dfforecasts["samp_incr"].values / dfforecasts["param2"].values)

            def fdist(param1, param2):
                """
                gamma.rvs(a=param1, scale=param2, size=1, random_state=None)
                """
                return(prng.gamma(param1, param2))
        else:
            raise ValueError("Invalid procdist specification: `{}`".format(procdist))

        dfforecasts["final_incr"] = np.where(
            dfforecasts["rectype"].values == "forecast",
            fdist(dfforecasts["param1"].values, dfforecasts["param2"].values) * dfforecasts["sign"].values,
            dfforecasts["samp_incr"].values
            )
        dfforecasts["final_cum"] = dfforecasts.groupby(["sim", "origin"])["final_incr"].cumsum()
        dfforecasts = dfforecasts.rename({"final_cum": "ultimate", "l_act_cum": "latest"}, axis=1)
        return(dfforecasts.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))


    @staticmethod
    def _bs_reserves(dfprocerror):
        """
        Compute unpaid loss reserve estimate using output from
        ``self._bs_process_error``.

        Parameters
        ----------
        dfprocerror: pd.DataFrame
            Output from ``self._bs_process_error``.

        Returns
        -------
        pd.DataFrame
        """
        keepcols = ["sim", "origin", "latest", "ultimate", "reserve"]
        max_devp = dfprocerror["dev"].values.max()
        dfprocerror["reserve"] = dfprocerror["ultimate"] - dfprocerror["latest"]
        dfreserves = dfprocerror[dfprocerror["dev"].values == max_devp][keepcols].drop_duplicates()
        dfreserves["latest"] = np.where(
            np.isnan(dfreserves["latest"].values),
            dfreserves["ultimate"].values, dfreserves["latest"].values
            )
        dfreserves["reserve"] = np.nan_to_num(dfreserves["reserve"].values, 0)
        return(dfreserves.sort_values(by=["origin", "sim"]).reset_index(drop=True))


    @staticmethod
    def _bs_std_error(dfreserves):
        """
        Compute standard error of bootstrapped reserves by origin and in aggregate.

        Parameters
        ----------
        dfreserves: pd.DataFrame
            Output from ``self._bs_reserves``.

        Returns
        -------
        pd.Series
        """
        # Compute standard deviation of bootstrap samples by origin.
        dforigin_std = dfreserves.groupby(["origin"], as_index=False)["reserve"].std(ddof=1)
        origin_se = pd.Series(
            data=dforigin_std["reserve"].values, index=dforigin_std["origin"].values,
            name="std_error")
        dftotal = dfreserves.groupby(["sim"], as_index=False)["reserve"].sum()
        total_se = pd.Series(
            data=dftotal["reserve"].std(ddof=1), index=["total"], name="std_error"
            )
        return(pd.concat([origin_se, total_se]))



class BootstrapChainLadderResult(BaseRangeEstimatorResult):
    """
    Container class for ``BootstrapChainLadder`` output.

    Parameters
    ----------
    summary: pd.DataFrame
        Chain Ladder summary compilation.

    reserve_dist: pd.DataFrame
        The predicitive distribution of reserve estimates generated via
        bootstrapping. ``reserve_dist`` is a five column DataFrame
        consisting of the simulation number, origin period, the latest
        loss amount for the associated origin period, and the predictive
        distribution of ultimates and reserves.

    sims_data: pd.DataFrame
        A DataFrame consiting of all simulated values an intermediate
        fields. When a large number of bootstrap iterations are run,
        ``sims_data`` will be correspondingly large. The fields include:

            - dev:
                The simulated development period.

            - incr:
                The actual incremental loss amount obtain from the fitted triangle.

            - incr_sqrt:
                The square root of incr.

            - l_act_cum:
                The latest actual cumulative loss amount for dev/origin.

            - l_act_dev:
                The latest dev period with actual losses for a given origin period.

            - ldf:
                Loss development factors computed on syntehtic triangle data.

            - origin:
                The simulated origin period.

            - rectype:
                Whether the dev/origin combination represent actual or forecast data
                in the squared triangle.

            - resid:
                The resampled adjusted residuals if ``parametric=False``, otherwise a
                random sampling from a normal distribution with mean zero and variance
                based on the variance of the adjusted residuals.

            - samp_cum:
                A syntehtic cumulative loss amount.

            - samp_incr:
                A synthetic incremental loss amount.

            - sim:
                Bootstrap iteration.

            - var:
                The variance, computed as scale_param x samp_incr.

            - sign:
                The sign of samp_incr.

            - param2/param1:
                Parameters for the process error distribution.

            - final_incr:
                Final simulated incremetnal loss amount after the incorporation of
                process error.

            - final_cum:
                Final simulated cumulative loss amount after the incorporation of
                process error.

    tri: trikit.triangle.CumTriangle
        A cumulative triangle instance.

    ldfs: pd.Series
        Loss development factors.

    scale_param: float
        The the sum of the squared unscaled Pearson residuals over the triangle's
        degrees of freedom.

    dof: int
        Triangle degrees of freedom.

    unscaled_residuals: pd.DataFrame
        The unscaled residuals.

    adjusted_residuals: pd.DataFrame
        The adjusted residuals.

    sampling_dist: np.ndarray
        Same as ``adjusted_residuals`` but as a numpy array with NaN's and 0's
        removed. None if ``parametric=True``.

    fitted_tri_cum: pd.DataFrame
        Cumulative triangle fit using backwards recursion.

    fitted_tri_incr: pd.DataFrame
        Incremental triangle fit using backwards recursion.

    sims: int
        Number of bootstrap iterations performed.

    procdist: str
        Distribution used to incorporate process variance. Currently "gamma" is
        the only option.

    parametric: bool
        Whether parametric or non-parametric bootstrap was performed.

    q: float or array_like of float
        Quantiles over which to evaluate reserve distribution in summary output.

    interpolation: {"linear", "lower", "higher", "midpoint", "nearest"}
        Optional parameter which specifies the interpolation method to use
        when the desired quantile lies between two data points i < j. See
        ``numpy.quantile`` for more information. Default value is "linear".

    kwargs: dict
        Additional keyword arguments passed into ``BootstrapChainLadder``'s
        ``__call__`` method.
    """
    def __init__(self, summary, tri, ldfs, tail, trisqrd, reserve_dist, sims_data,
                 scale_param, dof, unscaled_residuals, adjusted_residuals,
                 sampling_dist, fitted_tri_cum, fitted_tri_incr, sims, procdist,
                 parametric, q, interpolation, **kwargs):

        super().__init__(summary=summary, tri=tri, ldfs=ldfs, tail=tail,
                         trisqrd=trisqrd, process_error=None, parameter_error=None)

        self.unscaled_residuals = unscaled_residuals
        self.adjusted_residuals = adjusted_residuals
        self.fitted_tri_incr = fitted_tri_incr
        self.fitted_tri_cum = fitted_tri_cum
        self.sampling_dist = sampling_dist
        self.interpolation = interpolation
        self.reserve_dist = reserve_dist
        self.scale_param = scale_param
        self.parametric = parametric
        self.sims_data = sims_data
        self.procdist = procdist
        self.sims = sims
        self.dof = dof
        self.q = q

        if kwargs is not None:
            for kk in kwargs:
                setattr(self, kk, kwargs[kk])

        qtlsfields = [ii for ii in self.summary.columns if ii.endswith("%")]
        self.qtlhdrs = {ii: "{:,.0f}".format for ii in qtlsfields}
        self._summspecs.update(self.qtlhdrs)

        # Properties.
        self._residuals_detail = None
        self._fit_assessment = None
        self._origin_dist = None
        self._agg_dist = None


    @property
    def origin_dist(self):
        """
        Return distribution of bootstrapped ultimates/reserves by origin period.

        Returns
        -------
        pd.DataFrame
        """
        if self._origin_dist is None:
            dist_columns = ["latest", "ultimate", "reserve"]
            self._origin_dist = self.reserve_dist.groupby(
                ["sim", "origin"], as_index=False)[dist_columns].sum()
        return(self._origin_dist)


    @property
    def residuals_detail(self):
        """
        Summary statistics based on triangle residuals.

        Returns
        -------
        pd.DataFrame
        """
        if self._residuals_detail is None:
            if not self.parametric:
                unscaled = self.unscaled_residuals.values.ravel()
                adjusted = self.adjusted_residuals.values.ravel()
                unscaled = unscaled[~np.isnan(unscaled)]
                adjusted = adjusted[~np.isnan(adjusted)]
                unscaled = unscaled[unscaled != 0]
                adjusted = adjusted[adjusted != 0]
                unscaled_size = unscaled.size
                unscaled_sum = unscaled.sum(axis=0)
                unscaled_ssqr = np.sum(unscaled**2, axis=0)
                unscaled_min = unscaled.min(axis=0)
                unscaled_max = unscaled.max(axis=0)
                unscaled_mean = unscaled.mean(axis=0)
                unscaled_skew = stats.skew(unscaled, axis=0, nan_policy="omit")
                unscaled_mode = stats.mode(unscaled, axis=0, nan_policy="omit").mode[0]
                unscaled_cvar = stats.variation(unscaled, axis=0, nan_policy="omit")
                unscaled_kurt = stats.kurtosis(unscaled, axis=0, nan_policy="omit")
                unscaled_var = unscaled.var(ddof=1, axis=0)
                unscaled_std = unscaled.std(ddof=1, axis=0)
                unscaled_med = np.median(unscaled, axis=0)
                adjusted_size = adjusted.size
                adjusted_sum = adjusted.sum(axis=0)
                adjusted_ssqr = np.sum(adjusted**2, axis=0)
                adjusted_min = adjusted.min(axis=0)
                adjusted_max = adjusted.max(axis=0)
                adjusted_mean = adjusted.mean(axis=0)
                adjusted_skew = stats.skew(adjusted, axis=0, nan_policy="omit")
                adjusted_mode = stats.mode(adjusted, axis=0, nan_policy="omit").mode[0]
                adjusted_cvar = stats.variation(adjusted, axis=0, nan_policy="omit")
                adjusted_kurt = stats.kurtosis(adjusted, axis=0, nan_policy="omit")
                adjusted_var = adjusted.var(ddof=1, axis=0)
                adjusted_std = adjusted.std(ddof=1, axis=0)
                adjusted_med = np.median(adjusted, axis=0)
                self._residuals_detail = pd.DataFrame({
                    "unscaled": [
                        unscaled_size, unscaled_sum , unscaled_ssqr, unscaled_min,
                        unscaled_max, unscaled_mean, unscaled_skew, unscaled_mode,
                        unscaled_cvar, unscaled_kurt, unscaled_var , unscaled_std,
                        unscaled_med
                        ],
                    "adjusted": [
                        adjusted_size, adjusted_sum , adjusted_ssqr, adjusted_min,
                        adjusted_max, adjusted_mean, adjusted_skew, adjusted_mode,
                        adjusted_cvar, adjusted_kurt, adjusted_var , adjusted_std,
                        adjusted_med
                        ],
                    },
                    index=[
                        "size", "sum", "sum_of_squares", "minimum", "maximum", "mean",
                        "skew", "mode", "cov", "kurtosis", "variance",
                        "standard_deviation", "median"
                        ]
                    )

        return(self._residuals_detail)


    def _bs_data_transform(self, qtls, qtlhdrs):
        """
        Starts with BaseChainLadderResult's ``_data_transform``, and performs additional
        pre-processing in order to generate plot of bootstrapped reserve ranges by
        origin period.

        Returns
        -------
        pd.DataFrame
        """
        data0 = self._data_transform()
        data0 = data0[data0["origin"] != "total"]
        data1 = self._get_quantiles_by_devp(qtls, qtlhdrs)
        data1 = data1[data1["origin"] != "total"]
        data = data0.merge(data1, on=["origin", "dev"], how="left")

        # Remove qtlhdrs values where rectype=="actual".
        for qtlhdr in qtlhdrs:
            data[qtlhdr] = np.where(
                data["rectype"].values == "actual", np.NaN, data[qtlhdr].values
                )

        # Determine the first forecast period by origin, and set q-fields to actuals.
        increment = np.unique(self.ldfs.index[1:] - self.ldfs.index[:-1])[0]
        data["_ff"] = np.where(
            data["rectype"].values == "forecast",
            data["dev"].values, data["dev"].values.max() + increment
            )
        data["_minf"] = data.groupby(["origin"])["_ff"].transform("min")
        for hdr in qtlhdrs:
            data[hdr] = np.where(
                np.logical_and(
                    data["rectype"].values == "forecast",
                    data["_minf"].values == data["dev"].values
                    ), data["loss"].values, data[hdr].values
                )

        data = data.drop(["_ff", "_minf"], axis=1).reset_index(drop=True)
        dfv = data[["origin", "dev", "rectype", "loss"]]
        dfl = data[["origin", "dev", "rectype", qtlhdrs[0]]]
        dfu = data[["origin", "dev", "rectype", qtlhdrs[-1]]]
        dfl["rectype"] = qtlhdrs[0]
        dfl = dfl.rename({qtlhdrs[0]: "loss"}, axis=1)
        dfu["rectype"] = qtlhdrs[-1]
        dfu = dfu.rename({qtlhdrs[-1]: "loss"}, axis=1)
        return(pd.concat([dfv, dfl, dfu]).sort_index().reset_index(drop=True))


    def _get_quantiles_by_devp(self, qtls, qtlhdrs):
        """
        Get quantile of boostrapped reserve distribution for an individual origin
        period and in total.

        Parameters
        ----------
        q: array_like
            A length-2 sequence representing to upper and lower bounds of
            the estimated reserve distribution.

        Returns
        -------
        pd.DataFrame
        """
        dfsims = self.sims_data[["origin", "dev", "ultimate"]]
        dfults = dfsims[dfsims.dev == dfsims.dev.max()].reset_index(drop=True)
        dev_increment = np.unique(self.ldfs.index[1:] - self.ldfs.index[:-1])[0]
        dfults["dev"] = self.ldfs.index.max() + dev_increment
        dfsims = pd.concat([dfsims, dfults])
        dftotal_keys = dfsims[dfsims.origin == dfsims.origin.min()][["origin", "dev"]].drop_duplicates()
        dftotal_keys["origin"] = "total"
        dfqtls_keys = pd.concat(
            [dfsims[["origin", "dev"]].drop_duplicates(), dftotal_keys]
            ).reset_index(drop=True
            )

        # Get total reserve across all origin periods.
        dftotal = dfsims.copy(deep=True)
        dftotal["origin"] = "total"
        dftotal = dftotal.groupby(["origin", "dev"], as_index=False)

        dflist = []
        for ii, jj in zip(qtls, qtlhdrs):
            dfqtl = dfsims.groupby(["origin", "dev"], as_index=False).aggregate(
                "quantile", q=ii, method="linear").rename(
                {"ultimate": jj}, axis=1
                )
            dftotal_qtl = dftotal.aggregate(
                "quantile", q=ii, method="linear").rename({"ultimate": jj},
                axis=1
                )
            dflist.append(pd.concat([dfqtl, dftotal_qtl]))

        # Combine DataFrames in dflist into single table.
        dfqtls = functools.reduce(
            lambda df1, df2: df1.merge(df2, on=["origin", "dev"], how="left"),
            dflist, dfqtls_keys).reset_index(drop=True)
        return(dfqtls)


    def get_quantiles(self, q, interpolation="linear", lb=None):
        """
        Get quantiles of bootstrapped reserve distribution for an individual origin
        periods and in total. Returns a DataFrame, with columns representing the
        percentiles of interest.

        Parameters
        ----------
        q: array_like of float or float
            Quantile or sequence of quantiles to compute, which must be between 0 and 1
            inclusive.

        interpolation: {"linear", "lower", "higher", "midpoint", "nearest"}
            Optional parameter which specifies the interpolation method to use
            when the desired quantile lies between two data points i < j. See
            ``numpy.quantile`` for more information. Default value is "linear".

        lb: float
            Lower bound of simulated values. If ``lb`` is not None, quantiles less
            than ``lb`` will be set to ``lb``. To eliminate negative quantiles,
            set ``lb=0``.

        Returns
        -------
        pd.DataFrame
        """
        qarr = np.asarray(q, dtype=float)
        if np.any(np.logical_and(qarr > 1, qarr < 0)):
            raise ValueError("q values must fall within [0, 1].")
        else:
            qtls, qtlhdrs = self._qtls_formatter(q=q)
            qtl_pairs = [(qtlhdrs[ii], qtls[ii]) for ii in range(len(qtls))]
            dqq = {
                str(ii[0]): [
                    np.percentile(
                        self.reserve_dist[self.reserve_dist.origin == origin]["reserve"].values,
                        100 * ii[-1], method=interpolation
                        ) for origin in self.summary.index] for ii in qtl_pairs
                }
            dfqq = pd.DataFrame().from_dict(dqq).set_index(self.summary.index)
            if lb is not None:
                dfqq = dfqq.applymap(lambda v: lb if v < lb else v)
        return(dfqq)


    def plot(self, q=.90, actuals_color="#334488", forecasts_color="#FFFFFF",
             fill_color="#FCFCB1", fill_alpha=.75, axes_style="darkgrid",
             context="notebook", col_wrap=4, hue_kws=None, exhibit_path=None,
             **kwargs):
        """
        Generate exhibit representing the distribution of reserve estimates
        resulting from bootstrap resampling, along with percentiles from the
        distribution given by ``q``, the percentile(s) of interest.

        Parameters
        ----------
        q: float in range of [0,1]
            two_sided percentile interval to highlight, which must be between
            0 and 1 inclusive. For example, when ``q=.90``, the 5th and
            95th percentile of the ultimate/reserve distribution will be
            highlighted in the exhibit :math:`(\\frac{1 - q}{2}, \\frac{1 + q}{2})`.

        actuals_color: str
            Color or hexidecimal color code used to represent actuals.
            Defaults to "#00264C".

        forecasts_color: str
            Color or hexidecimal color code used to represent forecasts.
            Defaults to "#FFFFFF".

        fill_color: str
            Color or hexidecimal color code used to represent the fill color
            between reserve distribution quantiles associated with ``q``.
            Defaults to "#FCFCB1".

        fill_alpha: float
            Control transparency of ``fill_color`` between upper and lower
            percentile bounds of the ultimate/reserve distribution. Defaults
            to .75.

        axes_style: {"darkgrid", "whitegrid", "dark", "white", "ticks"}
            Aesthetic style of seaborn plots. Default values is "darkgrid".

        context: {"notebook", "paper", "talk", "poster"}.
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Default value is "notebook".

        col_wrap: int
            The maximum number of origin period axes to have on a single row
            of the resulting FacetGrid. Defaults to 5.

        hue_kws: dictionary of param:list of values mapping
            Other keyword arguments to insert into the plotting call to let
            other plot attributes vary across levels of the hue variable
            (e.g. the markers in a scatterplot). Each list of values should
            have length 4, with each index representing aesthetic
            overrides for forecasts, actuals, lower percentile and upper
            percentile renderings respectively. Defaults to ``None``.

        exhibit_path: str
            Path to which exhibit should be written. If None, exhibit will be
            rendered via ``plt.show()``.

        kwargs: dict
            Additional styling options for scatter points. This should include
            additional options accepted by ``plt.plot``.
        """
        import matplotlib
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set_context(context)

        qtls, qtlhdrs = self._qtls_formatter(q=q, two_sided=True)
        data = self._bs_data_transform(qtls, qtlhdrs).dropna(how="any", subset=["loss"])

        with sns.axes_style(axes_style):
            huekwargs = dict(
                marker=["o", "o", None, None], markersize=[6, 6, None, None],
                color=["#000000", "#000000", "#000000", "#000000"],
                fillstyle=["full", "full", "none", "none"],
                markerfacecolor=[forecasts_color, actuals_color, None, None],
                markeredgecolor=["#000000", "#000000", None, None],
                markeredgewidth=[.50, .50, None, None],
                linestyle=["-", "-", "-.", "--"], linewidth=[.475, .475, .625, .625],
                )

            if hue_kws is not None:
                # Determine whether the length of each element of hue_kws is 4.
                if all(len(hue_kws[i]) == 4 for i in hue_kws):
                    huekwargs.update(hue_kws)
                else:
                    warnings.warn("hue_kws overrides not correct length - Ignoring.")

            grid = sns.FacetGrid(
                data, col="origin", hue="rectype", hue_kws=huekwargs,
                col_wrap=col_wrap, margin_titles=False, despine=True,
                sharex=False, sharey=False,
                hue_order=["forecast", "actual", qtlhdrs[0], qtlhdrs[-1]]
                )

            ult_vals = grid.map(plt.plot, "dev", "loss",)
            devp_xticks = np.sort(data.dev.unique())
            devp_xticks_str = [
                str(ii) if ii != devp_xticks.max() else "ult" for ii in devp_xticks
                ]
            grid.set(xticks=devp_xticks)
            grid.set_xticklabels(devp_xticks_str, size=7)

            with warnings.catch_warnings():

                warnings.simplefilter("ignore")

                # Change ticklabel font size and place legend on each facet.
                for origin, ax_ii in zip(np.sort(data.origin.unique()), grid.axes):

                    legend = ax_ii.legend(
                        loc="lower right", fontsize="small", frameon=True,
                        fancybox=True, shadow=False, edgecolor="#909090",
                        framealpha=1, markerfirst=True,)
                    legend.get_frame().set_facecolor("#FFFFFF")

                    # For given origin, determine optimal 5-point tick labels.
                    origin_max_val = data[data.origin == origin].loss.max()
                    y_ticks, y_ticklabels = self._get_yticks(origin_max_val)
                    ax_ii.set_yticks(y_ticks)
                    ax_ii.set_yticklabels(y_ticklabels, size=7)

                    ax_ii.annotate(
                        origin, xy=(.075, .90), xytext=(.075, .90), xycoords='axes fraction',
                        textcoords='axes fraction', fontsize=9, rotation=0, color="#000000",
                        )

                    ax_ii.set_title("")
                    ax_ii.set_xlabel("")
                    ax_ii.set_ylabel("")

                    # Fill between upper and lower range bounds.
                    axc = ax_ii.get_children()
                    lines_ = [jj for jj in axc if isinstance(jj, matplotlib.lines.Line2D)]
                    xx = [jj._x for jj in lines_ if len(jj._x) > 0]
                    yy = [jj._y for jj in lines_ if len(jj._y) > 0]
                    x_, lb_, ub_ = xx[0], yy[-2], yy[-1]
                    ax_ii.fill_between(x_, lb_, ub_, color=fill_color, alpha=fill_alpha)

                    # Draw border around each facet.
                    for _, spine_ in ax_ii.spines.items():
                        spine_.set(visible=True, color="#000000", linewidth=.50)

            if exhibit_path is not None:
                plt.savefig(exhibit_path)
            else:
                plt.show()


    def hist(self, color="#FFFFFF", axes_style="darkgrid", context="notebook",
             col_wrap=4, exhibit_path=None, **kwargs):
        """
        Generate histogram of estimated reserve distribution by accident
        year and in total.

        Parameters
        ----------
        color: str
            Determines histogram color in each facet. Can also be specified as
            a key-value pair in ``kwargs``.
        axes_style: str
            Aesthetic style of plots. Defaults to "darkgrid". Other options
            include {whitegrid, dark, white, ticks}.

        context: str
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Defaults to ``"notebook"``. Additional options include
            {paper, talk, poster}.

        col_wrap: int
            The maximum number of origin period axes to have on a single row
            of the resulting FacetGrid. Defaults to 5.

        exhibit_path: str
            Path to which exhibit should be written. If None, exhibit will be
            rendered via ``plt.show()``.

        kwargs: dict
            Dictionary of optional matplotlib styling parameters.

        """
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set_context(context)

        # data0 = self.sims_data[["sim", "origin", "dev", "rectype", "latest", "reserve",]]
        # data0 = data0[(data0["dev"]==data0["dev"].max()) & (data0["rectype"]=="forecast")].reset_index(drop=True)
        # data0 = data0.drop(["dev", "rectype", "latest"], axis=1)
        #
        # # Include additional origin representing aggregate distribution.
        # data1 = data0.groupby("sim", as_index=False)[["reserve"]].sum()
        # data1["origin"] ="total"
        # data = pd.concat([data0, data1])
        data = self.reserve_dist

        # Get mean, min and max ultimate and reserve by origin.
        med_data = data.groupby("origin", as_index=False)[["reserve"]].median().rename(
            {"reserve": "med_res"}, axis=1).set_index("origin")
        min_data = data.groupby("origin", as_index=False)[["reserve"]].min().rename(
            {"reserve": "min_res"}, axis=1).set_index("origin")
        max_data = data.groupby("origin", as_index=False)[["reserve"]].max().rename(
            {"reserve": "max_res"}, axis=1).set_index("origin")
        dfmetrics = functools.reduce(lambda df1, df2: df1.join(df2), (med_data, min_data, max_data))
        dfmetrics = dfmetrics.applymap(lambda v: 0 if v < 0 else v).reset_index(drop=False)

        with sns.axes_style(axes_style):

            pltkwargs = {"color": color, "bins": 20, "edgecolor": "#484848",
                         "alpha": 1., "linewidth": .45}

            if kwargs is not None:
                pltkwargs.update(kwargs)

            grid = sns.FacetGrid(
                data, col="origin", col_wrap=col_wrap, margin_titles=False,
                despine=True, sharex=False, sharey=False,
                )

            hists = grid.map(plt.hist, "reserve", **pltkwargs)
            grid.set_axis_labels("", "")
            grid.set_titles("", size=6)

            # Change ticklabel font size and place legend on each facet.
            origin_vals = sorted([int(ii) for ii in data["origin"].unique() if ii != "total"])
            dindex = {jj: ii for ii, jj in enumerate(origin_vals)}
            dindex.update({"total": max(dindex.values()) + 1})
            data["origin_index"] = data["origin"].map(dindex)
            origin_order = data[["origin_index", "origin"]].drop_duplicates().sort_values(
                "origin_index"
                ).origin.values

            with warnings.catch_warnings():

                warnings.simplefilter("ignore")

                for origin, ax_ii in zip(origin_order, grid.axes):

                    # xmin = np.max([0, dfmetrics[dfmetrics.origin == origin]["min_res"].item()])
                    xmax = dfmetrics[dfmetrics.origin == origin]["max_res"].item() * 1.025
                    xmed = dfmetrics[dfmetrics.origin == origin]["med_res"].item()
                    origin_str = "{}".format(origin)
                    ax_ii.set_xlim([0, xmax])
                    ax_ii.axvline(xmed, color="#E02C70", linestyle="--", linewidth=1.5)
                    ax_ii.grid(False)

                    ymedloc = max(rect.get_height() for rect in ax_ii.patches) * .30
                    ax_ii.set_yticks([])
                    ax_ii.set_yticklabels([])
                    ax_ii.tick_params(
                        axis="x", which="both", bottom=True, top=False, labelbottom=True
                        )
                    ax_ii.set_xticklabels(
                        ["{:,.0f}".format(jj) for jj in ax_ii.get_xticks()], size=7
                        )
                    ax_ii.annotate(
                        origin_str, xy=(.85, .925), xycoords='axes fraction',
                        textcoords='axes fraction', fontsize=9, rotation=0, color="#000000",
                        )
                    ax_ii.annotate(
                        "median = {:,.0f}".format(xmed), (xmed, ymedloc), xytext=(7.5, 0),
                        textcoords="offset points", ha="center", va="bottom", fontsize=7,
                        rotation=90, color="#000000"
                        )

                    # Draw border around each facet.
                    for _, spine in ax_ii.spines.items():
                        spine.set(visible=True, color="#000000", linewidth=.50)

            if exhibit_path is not None:
                plt.savefig(exhibit_path)
            else:
                plt.show()



"""
Various trikit utilities. Contains convenience functions in support of
the CAS Loss Reserving Database and other sample datasets.
"""
import pandas as pd
from . import triangle



def _load(dataset, tri_type=None, dataref=None):
    """
    Load the specified sample dataset. If ``tri_type`` is not None, return sample
    dataset as specified triangle (one of ``{"cum", "incr"}``).

    Parameters
    ----------
    dataset: str
        Specifies which sample dataset to load. The complete set of sample
        datasets can be obtained by calling ``get_datasets``.

    tri_type: ``None`` or {"incr", "cum"}
        If ``None``, data subset is returned as pd.DataFrame. Otherwise,
        return subset as either incremental or cumulative triangle type.
        Default value is None.

    dataref: str
        Location of dataset reference.

    Returns
    -------
    Either pd.DataFrame, trikit.triangle.IncrTriangle or trikit.triangle.CumTriangle.
    """
    if dataset not in dataref.keys():
        raise KeyError("Specified dataset does not exist: `{}`".format(dataset))
    else:
        data_path = dataref[dataset]
        loss_data = pd.read_csv(data_path, delimiter=",")
        loss_data = loss_data[["origin", "dev", "value"]].reset_index(drop=True)

        if tri_type is not None:
            if not tri_type.startswith(("c", "i")):
                raise ValueError("tri_type must be one of {{'cum', 'incr'}}, not `{}`.".format(tri_type))
            else:
                loss_data = triangle.totri(loss_data, tri_type=tri_type)

    return(loss_data)



def _load_lrdb(tri_type=None, loss_type="incurred", lob="comauto", grcode=1767,
               grname=None, train_only=True, dataref=None):
    """
    Load the CAS Loss Reserving Database subset of losses. Additional
    keyword arguments are used to subset the CAS Loss Reserving Database to the
    records of interest.
    Within the Loss Reserving Database, "loss_key" and "grname" uniquely
    partition losses into 100 record blocks if ``lower_right_ind``=True,
    otherwise losses are partitioned into 55 record blocks. All available
    combinations of "loss_key" and "grcode" (referred to as "specs")
    can be obtained by calling ``get_lrdb_specs``.
    Note that when ``tri_type`` is "cum" or "incr", the remaining subset
    of records after applying ``lob``, ``grcode`` and ``grname`` filters will
    be aggregated into a single triangle.

    Parameters
    ----------
    tri_type: ``None`` or {"incr", "cum"}
        If ``None``, lrdb subset is returned as pd.DataFrame. Otherwise,
        return subset as either incremental or cumulative triangle type.
        Default value is None.

    lob: str
        Specifies the line of business to return. Available options are
        ``['comauto', 'ppauto', 'wkcomp', 'medmal', 'prodliab', 'othliab']``.

    grcode: str
        NAIC company code including insurer groups and single insurers.
        For a full listing, call ``get_lrdb_specs``.

    grname: str
        NAIC company name (including insurer groups and single insurers).
        The complete mapping of available grcodes can be obtained by
        calling ``get_lrdb_specs``.

    loss_type: {"paid", "incurred"}
        Specifies which loss data to load. Can be one of "paid" or
        "incurred". Defaults to "incurred". Note that bulk losses
        have already been subtracted from schedule P incurred losses.
        Applies only when ``dataset``="lrdb", otherwise parameter is
        ignored.

    train_only: bool
        If True, the upper-left portion of the triangle will be returned.
        The upper-left portion of the triangle typically consists of
        actual loss experience. If False, the completed triangle, consisting
        of 100 observations is returned. Defaults to True. Applies only when
        ``dataset`` ="lrdb", otherwise parameter is ignored.

    dataref: str
        Location of dataset reference.

    Returns
    -------
    Either pd.DataFrame, trikit.triangle.IncrTriangle or trikit.triangle.CumTriangle.
    """
    if not loss_type.startswith(("p", "i")):
        raise ValueError("loss_type should be one of {{'paid', 'incurred'}}, not `{}`.".format(loss_type))
    elif loss_type.lower().startswith("i"):
        loss_field = "incrd_loss"
    elif loss_type.lower().startswith("p"):
        loss_field = "paid_loss"

    data_path = dataref["lrdb"]
    loss_data = pd.read_csv(data_path, delimiter=",")
    loss_data = loss_data[
        ["loss_key", "grcode", "grname", "origin", "dev", loss_field, "train_ind"]
        ]

    if lob is not None:
        if lob not in loss_data["loss_key"].unique():
            raise ValueError("`{}` is not a valid lob selection.".format(lob))
        else:
            loss_data = loss_data[loss_data.loss_key == lob].reset_index(drop=True)

    if grcode is not None:
        if grcode not in loss_data["grcode"].unique():
            raise ValueError("`{}` is not a valid grcode selection.".format(grcode))
        else:
            loss_data = loss_data[loss_data.grcode == grcode].reset_index(drop=True)

    if grname is not None:
        if grname not in loss_data["grname"].unique():
            raise ValueError("`{}` is not a valid grname selection.".format(grname))
        else:
            loss_data = loss_data[loss_data.grname == grname].reset_index(drop=True)

    if train_only:
        loss_data = loss_data[loss_data.train_ind == 1].reset_index(drop=True)

    loss_data = loss_data.rename({loss_field: "value"}, axis=1)
    loss_data = loss_data[["origin", "dev", "value"]].reset_index(drop=True)

    if tri_type is not None:
        if not tri_type.startswith(("c", "i")):
            raise ValueError("tri_type should be one of {{'cum', 'incr'}}, not `{}`.".format(tri_type))
        else:
            loss_data = triangle.totri(loss_data, tri_type=tri_type)

    return(loss_data)



def _get_datasets(dataref):
    """
    Generate a list containing the names of available sample datasets.

    Parameters
    ----------
    dataref: str
        Location of dataset reference.

    Returns
    -------
    list
        Names of available sample datasets.
    """
    return(sorted([ii for ii in dataref.keys() if ii != "lrdb"]))



def _get_lrdb_lobs(lrdb_path):
    """
    Return the unique "loss_key" entries present in the CAS Loss
    Reserving Database (lrdb).

    Parameters
    ----------
    lrdb_path: str
        Location of CAS loss reserving database.

    Returns
    -------
    list
    """
    lrdb = pd.read_csv(lrdb_path, sep=",")
    lrdb = lrdb["loss_key"].unique()
    return(sorted(lrdb.tolist()))



def _get_lrdb_specs(lrdb_path):
    """
    Return a DataFrame containing the unique combinations of "loss_key",
    "grname" and "grcode" from the CAS Loss Reserving Database (lrdb).

    Parameters
    ----------
    lrdb_path: str
        Location of CAS loss reserving database.

    Returns
    -------
    pd.DataFrame
    """
    fields = ["loss_key", "grcode", "grname"]
    lrdb = pd.read_csv(lrdb_path, sep=",")
    lrdb = lrdb[fields].drop_duplicates().reset_index(drop=True)
    lrdb = lrdb.sort_values(by=["loss_key", "grcode"])
    return(lrdb)



"""
Triangle benchmarks.
"""
import numpy as np
import pandas as pd
import trikit




class TriangleSuite:
    def setup(self):
        self.data_i = trikit.load(dataset="raa")
        self.tri_c = trikit.totri(data=self.data_i, tri_type="cumulative")
        self.tri_i = trikit.totri(data=self.data_i, tri_type="incr")
        self.dftri_cum = pd.DataFrame(self.tri_c)
        self.dftri_incr = pd.DataFrame(self.tri_i)
        self.data_c = self.tri_c.to_tbl()
        self.tri = self.tri_c

    def time_incr_tabular_to_incr(self):
        trikit.totri(data=self.data_i, tri_type="incremental", data_format="incr", data_shape="tabular")

    def time_incr_tabular_to_cum(self):
        trikit.totri(data=self.data_i, tri_type="cumulative", data_format="incr", data_shape="tabular")

    def time_cum_tabular_to_incr(self):
        trikit.totri(data=self.data_c, tri_type="incremental", data_format="cum", data_shape="tabular")

    def time_cum_tabular_to_cum(self):
        trikit.totri(data=self.data_c, tri_type="cumulative", data_format="cum", data_shape="tabular")

    def time_incr_trilike_to_incr(self):
        trikit.totri(data=self.dftri_incr, tri_type="incremental", data_format="incr", data_shape="triangle")

    def time_incr_trilike_to_cum(self):
        trikit.totri(data=self.dftri_incr, tri_type="cumulative", data_format="incr", data_shape="triangle")

    def time_cum_trilike_to_incr(self):
        trikit.totri(data=self.dftri_cum, tri_type="incremental", data_format="cum", data_shape="triangle")

    def time_cum_trilike_to_cum(self):
        trikit.totri(data=self.dftri_cum, tri_type="cumulative", data_format="cum", data_shape="triangle")

    def time_nbr_cells(self):
        self.tri.nbr_cells

    def time_triind(self):
        self.tri.triind

    def time_origins(self):
        self.tri.origins

    def time_devp(self):
        self.tri.devp

    def time_rlvi(self):
        self.tri.rlvi

    def time_clvi(self):
        self.tri.clvi

    def time_latest_by_origin(self):
        self.tri.latest_by_origin

    def time_latest_by_devp(self):
        self.tri.latest_by_devp

    def time_maturity(self):
        self.tri.maturity

    def time_diagonal_0(self):
        self.tri.diagonal(offset=0)

    def time_diagonal_1(self):
        self.tri.diagonal(offset=-1)

    def time_diagonal_k(self):
        self.tri.diagonal(offset=self.tri.devp.size-2)

    def time_totbl(self):
        self.tri.to_tbl()

    def time_to_incr(self):
        self.tri.to_incr()

    def time_a2a(self):
        self.tri.a2a

    def time_a2aind(self):
        self.tri.a2aind

    def time_a2a_avgs(self):
        self.tri.a2a_avgs()



"""
Generalized Linear Model reserve estimators.
"""
import numpy as np
import pandas as pd 
import statsmodels.api as sm
import statsmodels.formula.api as smf


class GLMEstimator:
	
	def __init__(self, tri):

		pass


class GLMEstimatorResult:
	
	def __init__(self, **kwargs):
		
		pass
